"""Labware state store tests.

DEPRECATED: Testing LabwareView independently of LabwareStore is no
longer helpful. Try to add new tests to test_labware_state.py, where they can be
tested together, treating LabwareState as a private implementation detail.
"""

from contextlib import nullcontext as does_not_raise
from datetime import datetime
from typing import Any, ContextManager, Dict, List, Optional, Union, cast

import pytest
from numpy import isclose

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware import load_definition
from opentrons_shared_data.labware.labware_definition import (
    AxisAlignedBoundingBox3D,
    Extents,
    LabwareDefinition,
    LabwareDefinition2,
    LabwareDefinition3,
    LabwareRole,
    Parameters2,
    Vector3D,
    labware_definition_type_adapter,
)
from opentrons_shared_data.labware.labware_definition import (
    Dimensions as LabwareDimensions,
)
from opentrons_shared_data.pipette.types import LabwareUri

from opentrons.protocol_engine import errors
from opentrons.protocol_engine.state._axis_aligned_bounding_box import (
    AxisAlignedBoundingBox3D as EngineAABB,
)
from opentrons.protocol_engine.state._move_types import EdgePathType
from opentrons.protocol_engine.state.labware import (
    LabwareLoadParams,
    LabwareState,
    LabwareView,
)
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    Dimensions,
    GripSpecs,
    LabwareLocation,
    LabwareOffset,
    LabwareOffsetVector,
    LegacyLabwareOffsetLocation,
    LoadedLabware,
    ModuleLocation,
    ModuleModel,
    OnAddressableAreaOffsetLocationSequenceComponent,
    OnLabwareLocation,
    OnModuleOffsetLocationSequenceComponent,
)
from opentrons.types import DeckSlotName, MountType, Point

plate = LoadedLabware(
    id="plate-id",
    loadName="plate-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-plate-uri",
    offsetId=None,
    displayName="Fancy Plate Name",
)

flex_tiprack = LoadedLabware(
    id="flex-tiprack-id",
    loadName="flex-tiprack-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-flex-tiprack-uri",
    offsetId=None,
    displayName="Flex Tiprack Name",
)

reservoir = LoadedLabware(
    id="reservoir-id",
    loadName="reservoir-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
    definitionUri="some-reservoir-uri",
    offsetId=None,
)

trash = LoadedLabware(
    id="trash-id",
    loadName="trash-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
    definitionUri="some-trash-uri",
    offsetId=None,
)

tube_rack = LoadedLabware(
    id="tube-rack-id",
    loadName="tube-rack-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-tube-rack-uri",
    offsetId=None,
)

tip_rack = LoadedLabware(
    id="tip-rack-id",
    loadName="tip-rack-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-tip-rack-uri",
    offsetId=None,
)

adapter_plate = LoadedLabware(
    id="adapter-plate-id",
    loadName="adapter-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-adapter-uri",
    offsetId=None,
)

tiprack_lid = LoadedLabware(
    id="tiprack-lid-id",
    loadName="tiprack-lid-load-name",
    location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
    definitionUri="some-tiprack-lid-uri",
    offsetId=None,
)


def get_labware_view(
    labware_by_id: Optional[Dict[str, LoadedLabware]] = None,
    labware_offsets_by_id: Optional[Dict[str, LabwareOffset]] = None,
    definitions_by_uri: Optional[Dict[str, LabwareDefinition]] = None,
    deck_definition: Optional[DeckDefinitionV5] = None,
) -> LabwareView:
    """Get a labware view test subject."""
    state = LabwareState(
        labware_by_id=labware_by_id or {},
        labware_offsets_by_id=labware_offsets_by_id or {},
        definitions_by_uri=definitions_by_uri or {},
        deck_definition=deck_definition or cast(DeckDefinitionV5, {"fake": True}),
    )

    return LabwareView(state=state)


def test_get_labware_data_bad_id() -> None:
    """get_labware_data_by_id should raise if labware ID doesn't exist."""
    subject = get_labware_view()

    with pytest.raises(errors.LabwareNotLoadedError):
        subject.get("asdfghjkl")


def test_get_labware_data_by_id() -> None:
    """It should retrieve labware data from the state."""
    subject = get_labware_view(labware_by_id={"plate-id": plate})

    assert subject.get("plate-id") == plate


def test_get_id_by_module() -> None:
    """Should return the labware id associated to the module."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="test-uri",
                location=ModuleLocation(moduleId="module-id"),
            )
        }
    )
    assert subject.get_id_by_module(module_id="module-id") == "labware-id"


def test_get_id_by_module_raises_error() -> None:
    """Should raise error that labware not found."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="test-uri",
                location=ModuleLocation(moduleId="module-id"),
            )
        }
    )
    with pytest.raises(errors.exceptions.LabwareNotLoadedOnModuleError):
        subject.get_id_by_module(module_id="no-module-id")


def test_get_id_by_labware() -> None:
    """Should return the labware id associated to the labware."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="test-uri",
                location=OnLabwareLocation(labwareId="other-labware-id"),
            )
        }
    )
    assert subject.get_id_by_labware(labware_id="other-labware-id") == "labware-id"


def test_get_id_by_labware_raises_error() -> None:
    """Should raise an error that labware not found."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="test-uri",
                location=OnLabwareLocation(labwareId="other-labware-id"),
                displayName="The Labware",
            ),
        },
    )
    with pytest.raises(errors.exceptions.LabwareNotLoadedOnLabwareError):
        subject.get_id_by_labware(labware_id="labware-id")


def test_raise_if_labware_has_non_lid_labware_on_top() -> None:
    """It should raise if labware has a non-lid labware on top."""
    subject = get_labware_view(
        labware_by_id={
            "bottom-labware-1": LoadedLabware(
                id="bottom-labware-1",
                loadName="test",
                definitionUri="test-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                displayName="lorem",
            ),
            "bottom-labware-2": LoadedLabware(
                id="bottom-labware-2",
                loadName="test",
                definitionUri="test-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
                lid_id="lid-labware-a",
                displayName="ipsum",
            ),
            "bottom-labware-3": LoadedLabware(
                id="bottom-labware-3",
                loadName="test",
                definitionUri="test-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
                displayName="dolor",
            ),
            "lid-labware-a": LoadedLabware(
                id="lid-labware-a",
                loadName="lid",
                definitionUri="lid-uri",
                location=OnLabwareLocation(labwareId="bottom-labware-2"),
                displayName="sit",
            ),
            "top-labware-b": LoadedLabware(
                id="top-labware-b",
                loadName="test",
                definitionUri="test-uri",
                location=OnLabwareLocation(labwareId="bottom-labware-3"),
                displayName="amet",
            ),
        }
    )
    subject.raise_if_labware_has_non_lid_labware_on_top("bottom-labware-1")
    subject.raise_if_labware_has_non_lid_labware_on_top("bottom-labware-2")
    subject.raise_if_labware_has_non_lid_labware_on_top("lid-labware-a")
    subject.raise_if_labware_has_non_lid_labware_on_top("top-labware-b")
    with pytest.raises(errors.exceptions.LabwareIsInStackError):
        subject.raise_if_labware_has_non_lid_labware_on_top("bottom-labware-3")


def test_raise_if_labware_has_labware_on_top() -> None:
    """It should raise if labware has another labware on top, even if it's a lid."""
    subject = get_labware_view(
        labware_by_id={
            "bottom-labware-1": LoadedLabware(
                id="bottom-labware-1",
                loadName="test",
                definitionUri="test-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                displayName="lorem",
            ),
            "bottom-labware-2": LoadedLabware(
                id="bottom-labware-2",
                loadName="test",
                definitionUri="test-uri",
                location=ModuleLocation(moduleId="module-id"),
                lid_id="lid-labware-a",
                displayName="ipsum",
            ),
            "bottom-labware-3": LoadedLabware(
                id="bottom-labware-3",
                loadName="test",
                definitionUri="test-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
                displayName="dolor",
            ),
            "lid-labware-a": LoadedLabware(
                id="lid-labware-a",
                loadName="test-lid",
                definitionUri="lid-uri",
                location=OnLabwareLocation(labwareId="bottom-labware-2"),
                displayName="sit",
            ),
            "top-labware-b": LoadedLabware(
                id="top-labware-b",
                loadName="test",
                definitionUri="test-uri",
                location=OnLabwareLocation(labwareId="bottom-labware-3"),
                displayName="amet",
            ),
        }
    )
    subject.raise_if_labware_has_labware_on_top("bottom-labware-1")
    subject.raise_if_labware_has_labware_on_top("top-labware-b")
    subject.raise_if_labware_has_labware_on_top("lid-labware-a")
    with pytest.raises(errors.exceptions.LabwareIsInStackError):
        subject.raise_if_labware_has_labware_on_top("bottom-labware-2")
    with pytest.raises(errors.exceptions.LabwareIsInStackError):
        subject.raise_if_labware_has_labware_on_top("bottom-labware-3")


def test_get_labware_definition(well_plate_def: LabwareDefinition) -> None:
    """It should get a labware's definition from the state."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    assert subject.get_definition("plate-id") == well_plate_def


def test_get_labware_definition_bad_id() -> None:
    """get_labware_definition should raise if labware definition doesn't exist."""
    subject = get_labware_view()

    with pytest.raises(errors.LabwareDefinitionDoesNotExistError):
        subject.get_definition_by_uri(cast(LabwareUri, "not-a-uri"))


@pytest.mark.parametrize(
    argnames=["namespace", "version"],
    argvalues=[("world", 123), (None, 123), ("world", None), (None, None)],
)
def test_find_custom_labware_params(
    namespace: Optional[str], version: Optional[int]
) -> None:
    """It should find the missing (if any) load labware parameters."""
    labware_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(loadName="hello"),  # type: ignore[call-arg]
        namespace="world",
        version=123,
    )
    standard_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(loadName="goodbye"),  # type: ignore[call-arg]
        namespace="opentrons",
        version=456,
    )

    subject = get_labware_view(
        definitions_by_uri={
            "some-labware-uri": labware_def,
            "some-standard-uri": standard_def,
        },
    )

    result = subject.find_custom_labware_load_params()

    assert result == [
        LabwareLoadParams(load_name="hello", namespace="world", version=123)
    ]


def test_get_all_labware(
    well_plate_def: LabwareDefinition,
    reservoir_def: LabwareDefinition,
) -> None:
    """It should return all labware."""
    subject = get_labware_view(
        labware_by_id={
            "plate-id": plate,
            "reservoir-id": reservoir,
        }
    )

    all_labware = subject.get_all()

    assert all_labware == [plate, reservoir]


def test_get_labware_location() -> None:
    """It should return labware location."""
    subject = get_labware_view(labware_by_id={"plate-id": plate})

    result = subject.get_location("plate-id")

    assert result == DeckSlotLocation(slotName=DeckSlotName.SLOT_1)


@pytest.mark.parametrize(
    argnames="location",
    argvalues=[
        DeckSlotLocation(slotName=DeckSlotName.SLOT_D1),
        ModuleLocation(moduleId="module-id"),
        OFF_DECK_LOCATION,
    ],
)
def test_get_parent_location(location: LabwareLocation) -> None:
    """It should return the non-OnLabware location of a labware."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="plate-id",
                loadName="load-name",
                location=location,
                definitionUri="some-uri",
            )
        }
    )

    result = subject.get_parent_location(labware_id="labware-id")

    assert result == location


@pytest.mark.parametrize(
    argnames="location",
    argvalues=[
        DeckSlotLocation(slotName=DeckSlotName.SLOT_D1),
        ModuleLocation(moduleId="module-id"),
    ],
)
def test_get_parent_location_on_labware(location: LabwareLocation) -> None:
    """It should return the non-OnLabware location of a labware."""
    subject = get_labware_view(
        labware_by_id={
            "top-id": LoadedLabware(
                id="top-id",
                loadName="load-name",
                location=OnLabwareLocation(labwareId="middle-id"),
                definitionUri="some-uri",
            ),
            "middle-id": LoadedLabware(
                id="middle-id",
                loadName="load-name",
                location=OnLabwareLocation(labwareId="bottom-id"),
                definitionUri="some-uri",
            ),
            "bottom-id": LoadedLabware(
                id="bottom-id",
                loadName="load-name",
                location=location,
                definitionUri="some-uri",
            ),
        }
    )

    result = subject.get_parent_location(labware_id="top-id")

    assert result == location


def test_get_has_quirk(
    well_plate_def: LabwareDefinition,
    reservoir_def: LabwareDefinition,
) -> None:
    """It should return whether a labware by ID has a given quirk."""
    subject = get_labware_view(
        labware_by_id={
            "plate-id": plate,
            "reservoir-id": reservoir,
        },
        definitions_by_uri={
            "some-plate-uri": well_plate_def,
            "some-reservoir-uri": reservoir_def,
        },
    )

    well_plate_has_center_quirk = subject.get_has_quirk(
        labware_id="plate-id",
        quirk="centerMultichannelOnWells",
    )

    reservoir_has_center_quirk = subject.get_has_quirk(
        labware_id="reservoir-id",
        quirk="centerMultichannelOnWells",
    )

    assert well_plate_has_center_quirk is False
    assert reservoir_has_center_quirk is True


def test_quirks(
    well_plate_def: LabwareDefinition,
    reservoir_def: LabwareDefinition,
) -> None:
    """It should return a labware's quirks."""
    subject = get_labware_view(
        labware_by_id={
            "plate-id": plate,
            "reservoir-id": reservoir,
        },
        definitions_by_uri={
            "some-plate-uri": well_plate_def,
            "some-reservoir-uri": reservoir_def,
        },
    )

    well_plate_quirks = subject.get_quirks("plate-id")
    reservoir_quirks = subject.get_quirks("reservoir-id")

    assert well_plate_quirks == []
    assert reservoir_quirks == ["centerMultichannelOnWells", "touchTipDisabled"]


def test_get_well_definition_bad_name(well_plate_def: LabwareDefinition) -> None:
    """get_well_definition should raise if well name doesn't exist."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    with pytest.raises(errors.WellDoesNotExistError):
        subject.get_well_definition(labware_id="plate-id", well_name="foobar")


def test_get_well_definition(well_plate_def: LabwareDefinition) -> None:
    """It should return a well definition by well name."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    expected_well_def = well_plate_def.wells["B2"]
    result = subject.get_well_definition(labware_id="plate-id", well_name="B2")

    assert result == expected_well_def


def test_get_well_definition_get_first(well_plate_def: LabwareDefinition) -> None:
    """It should return the first well definition if no given well name."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    expected_well_def = well_plate_def.wells["A1"]
    result = subject.get_well_definition(labware_id="plate-id", well_name=None)

    assert result == expected_well_def


def test_get_well_geometry_raises_error(well_plate_def: LabwareDefinition) -> None:
    """It should raise an IncompleteLabwareDefinitionError when there's no innerLabwareGeometry."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    with pytest.raises(errors.IncompleteLabwareDefinitionError):
        subject.get_well_geometry(labware_id="plate-id")


def test_get_well_size_circular(well_plate_def: LabwareDefinition) -> None:
    """It should return the well dimensions of a circular well."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )
    expected_well_def = well_plate_def.wells["A2"]
    assert expected_well_def.shape == "circular"  # For type checking.
    expected_size = (
        expected_well_def.diameter,
        expected_well_def.diameter,
        expected_well_def.depth,
    )

    result = subject.get_well_size(labware_id="plate-id", well_name="A2")

    assert result == expected_size


def test_get_well_size_rectangular(reservoir_def: LabwareDefinition) -> None:
    """It should return the well dimensions of a rectangular well."""
    subject = get_labware_view(
        labware_by_id={"reservoir-id": reservoir},
        definitions_by_uri={"some-reservoir-uri": reservoir_def},
    )
    expected_well_def = reservoir_def.wells["A2"]
    assert expected_well_def.shape == "rectangular"  # For type checking.
    expected_size = (
        expected_well_def.xDimension,
        expected_well_def.yDimension,
        expected_well_def.depth,
    )

    result = subject.get_well_size(labware_id="reservoir-id", well_name="A2")

    assert result == expected_size


def test_labware_has_well(falcon_tuberack_def: LabwareDefinition) -> None:
    """It should return a list of wells from definition."""
    subject = get_labware_view(
        labware_by_id={"tube-rack-id": tube_rack},
        definitions_by_uri={"some-tube-rack-uri": falcon_tuberack_def},
    )

    result = subject.validate_liquid_allowed_in_labware(
        labware_id="tube-rack-id", wells={"A1": 30, "B1": 100}
    )
    assert result == ["A1", "B1"]

    with pytest.raises(errors.WellDoesNotExistError):
        subject.validate_liquid_allowed_in_labware(
            labware_id="tube-rack-id", wells={"AA": 30}
        )

    with pytest.raises(errors.LabwareNotLoadedError):
        subject.validate_liquid_allowed_in_labware(labware_id="no-id", wells={"A1": 30})


def test_validate_liquid_allowed_raises_incompatible_labware() -> None:
    """It should raise when validating labware that is a tiprack or an adapter."""
    subject = get_labware_view(
        labware_by_id={
            "tiprack-id": LoadedLabware(
                id="tiprack-id",
                loadName="test1",
                definitionUri="some-tiprack-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                displayName="Tip Rack",
            ),
            "adapter-id": LoadedLabware(
                id="adapter-id",
                loadName="test2",
                definitionUri="some-adapter-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
                displayName="Some adapter",
            ),
        },
        definitions_by_uri={
            "some-tiprack-uri": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(isTiprack=True),  # type: ignore[call-arg]
                wells={},
            ),
            "some-adapter-uri": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(isTiprack=False),  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                wells={},
            ),
        },
    )

    with pytest.raises(errors.LabwareIsTipRackError):
        subject.validate_liquid_allowed_in_labware(labware_id="tiprack-id", wells={})

    with pytest.raises(errors.LabwareIsAdapterError):
        subject.validate_liquid_allowed_in_labware(labware_id="adapter-id", wells={})


def test_get_tip_length_raises_with_non_tip_rack(
    well_plate_def: LabwareDefinition,
) -> None:
    """It should raise if you try to get the tip length of a regular labware."""
    subject = get_labware_view(
        labware_by_id={"plate-id": plate},
        definitions_by_uri={"some-plate-uri": well_plate_def},
    )

    with pytest.raises(errors.LabwareIsNotTipRackError):
        subject.get_tip_length("plate-id")


def test_get_tip_length_gets_length_from_definition(
    tip_rack_def: LabwareDefinition,
) -> None:
    """It should return the tip length from the definition."""
    subject = get_labware_view(
        labware_by_id={"tip-rack-id": tip_rack},
        definitions_by_uri={"some-tip-rack-uri": tip_rack_def},
    )

    length = subject.get_tip_length("tip-rack-id", 12.3)
    assert length == tip_rack_def.parameters.tipLength - 12.3  # type: ignore[operator]


def test_get_tip_drop_z_offset() -> None:
    """It should get a tip drop z offset by scaling the tip length."""
    tip_rack_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        parameters=Parameters2.model_construct(  # type: ignore[call-arg]
            tipLength=100,
        )
    )

    subject = get_labware_view(
        labware_by_id={"tip-rack-id": tip_rack},
        definitions_by_uri={"some-tip-rack-uri": tip_rack_def},
    )

    result = subject.get_tip_drop_z_offset(
        labware_id="tip-rack-id", length_scale=0.5, additional_offset=-0.123
    )

    assert result == -50.123


def test_get_labware_uri_from_definition(tip_rack_def: LabwareDefinition) -> None:
    """It should return the labware's definition URI."""
    tip_rack = LoadedLabware(
        id="tip-rack-id",
        loadName="tip-rack-load-name",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        definitionUri="some-tip-rack-uri",
        offsetId=None,
    )

    subject = get_labware_view(
        labware_by_id={"tip-rack-id": tip_rack},
        definitions_by_uri={"some-tip-rack-uri": tip_rack_def},
    )

    result = subject.get_definition_uri(labware_id="tip-rack-id")
    assert result == "some-tip-rack-uri"


def test_get_labware_uri_from_full_definition(tip_rack_def: LabwareDefinition) -> None:
    """It should be able to construct a URI given a full definition."""
    subject = get_labware_view()
    result = subject.get_uri_from_definition(tip_rack_def)
    assert result == "opentrons/opentrons_96_tiprack_300ul/1"


def test_is_tiprack(
    tip_rack_def: LabwareDefinition, reservoir_def: LabwareDefinition
) -> None:
    """It should determine if labware is a tip rack."""
    subject = get_labware_view(
        labware_by_id={
            "tip-rack-id": tip_rack,
            "reservoir-id": reservoir,
        },
        definitions_by_uri={
            "some-tip-rack-uri": tip_rack_def,
            "some-reservoir-uri": reservoir_def,
        },
    )

    assert subject.is_tiprack(labware_id="tip-rack-id") is True
    assert subject.is_tiprack(labware_id="reservoir-id") is False


def test_is_lid(
    reservoir_def: LabwareDefinition, tiprack_lid_def: LabwareDefinition
) -> None:
    """It should return True if labware is a lid."""
    subject = get_labware_view(
        labware_by_id={
            "reservoir-id": reservoir,
            "tiprack-lid-id": tiprack_lid,
        },
        definitions_by_uri={
            "some-reservoir-uri": reservoir_def,
            "some-tiprack-lid-uri": tiprack_lid_def,
        },
    )
    assert subject.is_lid(labware_id="reservoir-id") is False
    assert subject.is_lid(labware_id="tiprack-lid-id") is True


def test_get_load_name(reservoir_def: LabwareDefinition) -> None:
    """It should return the load name."""
    subject = get_labware_view(
        labware_by_id={"reservoir-id": reservoir},
        definitions_by_uri={"some-reservoir-uri": reservoir_def},
    )

    result = subject.get_load_name("reservoir-id")

    assert result == reservoir_def.parameters.loadName


def test_get_dimensions() -> None:
    """It should compute the dimensions of a labware."""
    subject = get_labware_view()

    # Schema 2 case:
    assert subject.get_dimensions(
        labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            schemaVersion=2,
            cornerOffsetFromSlot=Vector3D(
                x=1, y=2, z=3
            ),  # Should not affect dimensions.
            dimensions=LabwareDimensions(
                xDimension=100, yDimension=200, zDimension=300
            ),
        )
    ) == Dimensions(100, 200, 300)

    # Schema 3 case:
    assert subject.get_dimensions(
        labware_definition=LabwareDefinition3.model_construct(  # type: ignore[call-arg]
            schemaVersion=3,
            extents=Extents(
                total=AxisAlignedBoundingBox3D(
                    backLeftBottom=Vector3D(x=1, y=2, z=3),
                    frontRightTop=Vector3D(x=101, y=-198, z=303),
                )
            ),
        )
    ) == Dimensions(100, 200, 300)


def test_get_extents_around_lw_origin() -> None:
    """It should compute the extents of the labware, relative to the labware's origin."""
    subject = get_labware_view()

    # Schema 2 case:
    assert subject.get_extents_around_lw_origin(
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            schemaVersion=2,
            cornerOffsetFromSlot=Vector3D(x=1, y=2, z=3),  # Should not affect extents.
            dimensions=LabwareDimensions(
                xDimension=100, yDimension=200, zDimension=300
            ),
        )
    ) == EngineAABB.from_corners(Point(0, 0, 0), Point(100, 200, 300))

    # Schema 3 case:
    assert subject.get_extents_around_lw_origin(
        LabwareDefinition3.model_construct(  # type: ignore[call-arg]
            schemaVersion=3,
            extents=Extents(
                total=AxisAlignedBoundingBox3D(
                    backLeftBottom=Vector3D(x=100, y=200, z=300),
                    frontRightTop=Vector3D(x=50, y=250, z=350),
                )
            ),
        )
    ) == EngineAABB.from_corners(Point(100, 200, 300), Point(50, 250, 350))


def test_get_default_magnet_height(
    magdeck_well_plate_def: LabwareDefinition,
) -> None:
    """Should get get the default value for magnetic height."""
    well_plate = LoadedLabware(
        id="well-plate-id",
        loadName="load-name",
        location=ModuleLocation(moduleId="module-id"),
        definitionUri="well-plate-uri",
        offsetId=None,
    )

    subject = get_labware_view(
        labware_by_id={"well-plate-id": well_plate},
        definitions_by_uri={"well-plate-uri": magdeck_well_plate_def},
    )

    assert subject.get_default_magnet_height(module_id="module-id", offset=2) == 12.0


def test_get_deck_definition(ot2_standard_deck_def: DeckDefinitionV5) -> None:
    """It should get the deck definition from the state."""
    subject = get_labware_view(deck_definition=ot2_standard_deck_def)

    assert subject.get_deck_definition() == ot2_standard_deck_def


def test_get_labware_offset_vector() -> None:
    """It should get a labware's offset vector."""
    labware_without_offset = LoadedLabware(
        id="without-offset-labware-id",
        loadName="labware-load-name",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        definitionUri="some-labware-uri",
        offsetId=None,
    )

    labware_with_offset = LoadedLabware(
        id="with-offset-labware-id",
        loadName="labware-load-name",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        definitionUri="some-labware-uri",
        offsetId="offset-id",
    )

    offset_vector = LabwareOffsetVector(x=1, y=2, z=3)
    offset = LabwareOffset(
        id="offset-id",
        createdAt=datetime(year=2021, month=1, day=2),
        definitionUri="some-labware-uri",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=offset_vector,
    )

    subject = get_labware_view(
        labware_by_id={
            labware_without_offset.id: labware_without_offset,
            labware_with_offset.id: labware_with_offset,
        },
        labware_offsets_by_id={"offset-id": offset},
    )

    assert subject.get_labware_offset_vector(labware_with_offset.id) == offset.vector

    assert subject.get_labware_offset_vector(
        labware_without_offset.id
    ) == LabwareOffsetVector(x=0, y=0, z=0)

    with pytest.raises(errors.LabwareNotLoadedError):
        subject.get_labware_offset_vector("wrong-labware-id")


def test_get_labware_offset() -> None:
    """It should return the requested labware offset, if it exists."""
    offset_a = LabwareOffset(
        id="id-a",
        createdAt=datetime(year=2021, month=1, day=1),
        definitionUri="uri-a",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=1, z=1),
    )

    offset_b = LabwareOffset(
        id="id-b",
        createdAt=datetime(year=2022, month=2, day=2),
        definitionUri="uri-b",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_2),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="2")
        ],
        vector=LabwareOffsetVector(x=2, y=2, z=2),
    )

    subject = get_labware_view(
        labware_offsets_by_id={"id-a": offset_a, "id-b": offset_b}
    )

    assert subject.get_labware_offset("id-a") == offset_a
    assert subject.get_labware_offset("id-b") == offset_b
    with pytest.raises(errors.LabwareOffsetDoesNotExistError):
        subject.get_labware_offset("wrong-labware-offset-id")


def test_get_labware_offsets() -> None:
    """It should return a list of all labware offsets, in order."""
    offset_a = LabwareOffset(
        id="id-a",
        createdAt=datetime(year=2021, month=1, day=1),
        definitionUri="uri-a",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=1, z=1),
    )

    offset_b = LabwareOffset(
        id="id-b",
        createdAt=datetime(year=2022, month=2, day=2),
        definitionUri="uri-b",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_2),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="2")
        ],
        vector=LabwareOffsetVector(x=2, y=2, z=2),
    )

    empty_subject = get_labware_view()
    assert empty_subject.get_labware_offsets() == []

    filled_subject_a_before_b = get_labware_view(
        labware_offsets_by_id={"id-a": offset_a, "id-b": offset_b}
    )
    assert filled_subject_a_before_b.get_labware_offsets() == [offset_a, offset_b]

    filled_subject_b_before_a = get_labware_view(
        labware_offsets_by_id={"id-b": offset_b, "id-a": offset_a}
    )
    assert filled_subject_b_before_a.get_labware_offsets() == [offset_b, offset_a]


def test_find_applicable_labware_offset() -> None:
    """It should return the most recent offset with matching URI and location."""
    offset_1 = LabwareOffset(
        id="id-1",
        createdAt=datetime(year=2021, month=1, day=1),
        definitionUri="definition-uri",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=1, z=1),
    )

    # Same definitionUri and location; different id, createdAt, and offset.
    offset_2 = LabwareOffset(
        id="id-2",
        createdAt=datetime(year=2022, month=2, day=2),
        definitionUri="definition-uri",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=2, y=2, z=2),
    )

    offset_3 = LabwareOffset(
        id="id-3",
        createdAt=datetime(year=2023, month=3, day=3),
        definitionUri="on-module-definition-uri",
        location=LegacyLabwareOffsetLocation(
            slotName=DeckSlotName.SLOT_1,
            moduleModel=ModuleModel.TEMPERATURE_MODULE_V1,
        ),
        locationSequence=[
            OnModuleOffsetLocationSequenceComponent(
                moduleModel=ModuleModel.TEMPERATURE_MODULE_V1
            ),
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1"),
        ],
        vector=LabwareOffsetVector(x=3, y=3, z=3),
    )

    subject = get_labware_view(
        # Simulate offset_2 having been added after offset_1.
        labware_offsets_by_id={"id-1": offset_1, "id-2": offset_2, "id-3": offset_3}
    )

    # Matching both definitionURI and location. Should return 2nd (most recent) offset.
    assert (
        subject.find_applicable_labware_offset(
            definition_uri="definition-uri",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                )
            ],
        )
        == offset_2
    )

    assert (
        subject.find_applicable_labware_offset(
            definition_uri="on-module-definition-uri",
            location=[
                OnModuleOffsetLocationSequenceComponent(
                    moduleModel=ModuleModel.TEMPERATURE_MODULE_V1
                ),
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                ),
            ],
        )
        == offset_3
    )

    # Doesn't match anything, since definitionUri is different.
    assert (
        subject.find_applicable_labware_offset(
            definition_uri="different-definition-uri",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                )
            ],
        )
        is None
    )

    # Doesn't match anything, since location is different.
    assert (
        subject.find_applicable_labware_offset(
            definition_uri="different-definition-uri",
            location=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="2"
                )
            ],
        )
        is None
    )


def test_get_user_specified_display_name() -> None:
    """It should get a labware's user-specified display name."""
    subject = get_labware_view(
        labware_by_id={
            "plate_with_display_name": plate,
            "reservoir_without_display_name": reservoir,
        },
    )

    assert (
        subject.get_user_specified_display_name("plate_with_display_name")
        == "Fancy Plate Name"
    )
    assert (
        subject.get_user_specified_display_name("reservoir_without_display_name")
        is None
    )


def test_get_display_name(
    well_plate_def: LabwareDefinition,
    reservoir_def: LabwareDefinition,
) -> None:
    """It should get the labware's display name."""
    subject = get_labware_view(
        labware_by_id={
            "plate_with_custom_display_name": plate,
            "reservoir_with_default_display_name": reservoir,
        },
        definitions_by_uri={
            "some-plate-uri": well_plate_def,
            "some-reservoir-uri": reservoir_def,
        },
    )
    assert (
        subject.get_display_name("plate_with_custom_display_name") == "Fancy Plate Name"
    )
    assert (
        subject.get_display_name("reservoir_with_default_display_name")
        == "NEST 12 Well Reservoir 15 mL"
    )


def test_get_fixed_trash_id() -> None:
    """It should return the ID of the labware loaded into the fixed trash slot."""
    # OT-2 fixed trash slot:
    subject = get_labware_view(
        labware_by_id={
            "abc123": LoadedLabware(
                id="abc123",
                loadName="trash-load-name",
                location=DeckSlotLocation(slotName=DeckSlotName.FIXED_TRASH),
                definitionUri="trash-definition-uri",
                offsetId=None,
                displayName=None,
            )
        },
    )
    assert subject.get_fixed_trash_id() == "abc123"

    # OT-3 fixed trash slot:
    subject = get_labware_view(
        labware_by_id={
            "abc123": LoadedLabware(
                id="abc123",
                loadName="trash-load-name",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_A3),
                definitionUri="trash-definition-uri",
                offsetId=None,
                displayName=None,
            )
        },
    )
    assert subject.get_fixed_trash_id() == "abc123"

    # Nothing in the fixed trash slot:
    subject = get_labware_view(
        labware_by_id={
            "abc123": LoadedLabware(
                id="abc123",
                loadName="trash-load-name",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                definitionUri="trash-definition-uri",
                offsetId=None,
                displayName=None,
            )
        },
    )
    assert subject.get_fixed_trash_id() is None


@pytest.mark.parametrize(
    argnames=["location", "expected_raise"],
    argvalues=[
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            pytest.raises(errors.LocationIsOccupiedError),
        ),
        (
            ModuleLocation(moduleId="module-id"),
            pytest.raises(errors.LocationIsOccupiedError),
        ),
        (DeckSlotLocation(slotName=DeckSlotName.SLOT_2), does_not_raise()),
        (ModuleLocation(moduleId="non-matching-id"), does_not_raise()),
    ],
)
def test_raise_if_labware_in_location(
    location: Union[DeckSlotLocation, ModuleLocation],
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if there is labware in specified location."""
    subject = get_labware_view(
        labware_by_id={
            "abc123": LoadedLabware(
                id="abc123",
                loadName="labware-1",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
                definitionUri="labware-definition-uri",
                offsetId=None,
                displayName=None,
            ),
            "xyz456": LoadedLabware(
                id="xyz456",
                loadName="labware-2",
                location=ModuleLocation(moduleId="module-id"),
                definitionUri="labware-definition-uri",
                offsetId=None,
                displayName=None,
            ),
        }
    )
    with expected_raise:
        subject.raise_if_labware_in_location(location=location)


def test_get_by_slot() -> None:
    """It should get the labware in a given slot."""
    labware_1 = LoadedLabware.model_construct(  # type: ignore[call-arg]
        id="1", location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )
    labware_2 = LoadedLabware.model_construct(  # type: ignore[call-arg]
        id="2", location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2)
    )
    labware_3 = LoadedLabware.model_construct(  # type: ignore[call-arg]
        id="3", location=ModuleLocation(moduleId="cool-module")
    )

    subject = get_labware_view(
        labware_by_id={"1": labware_1, "2": labware_2, "3": labware_3}
    )

    assert subject.get_by_slot(DeckSlotName.SLOT_1) == labware_1
    assert subject.get_by_slot(DeckSlotName.SLOT_2) == labware_2
    assert subject.get_by_slot(DeckSlotName.SLOT_3) is None


@pytest.mark.parametrize(
    ["well_name", "mount", "labware_slot", "next_to_module", "expected_result"],
    [
        ("abc", MountType.RIGHT, DeckSlotName.SLOT_3, False, EdgePathType.LEFT),
        ("abc", MountType.RIGHT, DeckSlotName.SLOT_D3, False, EdgePathType.LEFT),
        ("abc", MountType.RIGHT, DeckSlotName.SLOT_1, True, EdgePathType.LEFT),
        ("abc", MountType.RIGHT, DeckSlotName.SLOT_D1, True, EdgePathType.LEFT),
        ("pqr", MountType.LEFT, DeckSlotName.SLOT_3, True, EdgePathType.RIGHT),
        ("pqr", MountType.LEFT, DeckSlotName.SLOT_D3, True, EdgePathType.RIGHT),
        ("pqr", MountType.LEFT, DeckSlotName.SLOT_3, False, EdgePathType.DEFAULT),
        ("pqr", MountType.LEFT, DeckSlotName.SLOT_D3, False, EdgePathType.DEFAULT),
        ("pqr", MountType.RIGHT, DeckSlotName.SLOT_3, True, EdgePathType.DEFAULT),
        ("pqr", MountType.RIGHT, DeckSlotName.SLOT_D3, True, EdgePathType.DEFAULT),
        ("def", MountType.LEFT, DeckSlotName.SLOT_3, True, EdgePathType.DEFAULT),
        ("def", MountType.LEFT, DeckSlotName.SLOT_D3, True, EdgePathType.DEFAULT),
    ],
)
def test_get_edge_path_type(
    well_name: str,
    mount: MountType,
    labware_slot: DeckSlotName,
    next_to_module: bool,
    expected_result: EdgePathType,
) -> None:
    """It should get the proper edge path type based on well name, mount, and labware position."""
    labware = LoadedLabware(
        id="tip-rack-id",
        loadName="load-name",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        definitionUri="some-labware-uri",
        offsetId=None,
    )

    labware_def = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        ordering=[["abc", "def"], ["ghi", "jkl"], ["mno", "pqr"]]
    )

    subject = get_labware_view(
        labware_by_id={"labware-id": labware},
        definitions_by_uri={
            "some-labware-uri": labware_def,
        },
    )

    result = subject.get_edge_path_type(
        "labware-id", well_name, mount, labware_slot, next_to_module
    )

    assert result == expected_result


def test_get_all_labware_definition(
    tip_rack_def: LabwareDefinition, falcon_tuberack_def: LabwareDefinition
) -> None:
    """It should return the loaded labware definition list."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="opentrons_96_tiprack_300ul",
                location=ModuleLocation(moduleId="module-id"),
            )
        },
        definitions_by_uri={
            "opentrons_96_tiprack_300ul": tip_rack_def,
            "falcon-definition": falcon_tuberack_def,
        },
    )

    result = subject.get_loaded_labware_definitions()

    assert result == [tip_rack_def]


def test_get_all_labware_definition_empty() -> None:
    """It should return an empty list."""
    subject = get_labware_view(
        labware_by_id={},
    )

    result = subject.get_loaded_labware_definitions()

    assert result == []


def test_raise_if_labware_cannot_be_stacked_is_adapter() -> None:
    """It should raise if the labware trying to be stacked is an adapter."""
    subject = get_labware_view()

    with pytest.raises(
        errors.LabwareCannotBeStackedError, match="defined as an adapter"
    ):
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(loadName="name"),  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
            ),
            bottom_labware_id="labware-id",
        )


def test_raise_if_labware_cannot_be_stacked_not_validated() -> None:
    """It should raise if the labware name is not in the definition stacking overlap."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="def-uri",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            )
        },
    )

    with pytest.raises(
        errors.LabwareCannotBeStackedError, match="loaded onto labware test"
    ):
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(loadName="name"),  # type: ignore[call-arg]
                stackingOffsetWithLabware={},
            ),
            bottom_labware_id="labware-id",
        )


def test_raise_if_labware_cannot_be_stacked_on_module_not_adapter() -> None:
    """It should raise if the below labware on a module is not an adapter."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="def-uri",
                location=ModuleLocation(moduleId="module-id"),
            )
        },
        definitions_by_uri={
            "def-uri": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware]
            )
        },
    )

    with pytest.raises(errors.LabwareCannotBeStackedError, match="module"):
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(loadName="name"),  # type: ignore[call-arg]
                stackingOffsetWithLabware={"test": Vector3D(x=0, y=0, z=0)},
            ),
            bottom_labware_id="labware-id",
        )


def test_raise_if_labware_cannot_be_stacked_on_labware_on_adapter() -> None:
    """It should raise if the OnLabware location is on an adapter."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="test",
                definitionUri="def-uri-1",
                location=OnLabwareLocation(labwareId="below-id"),
            ),
            "below-id": LoadedLabware(
                id="below-id",
                loadName="adapter-name",
                definitionUri="def-uri-2",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
        },
        definitions_by_uri={
            "def-uri-1": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(loadName="test"),  # type: ignore[call-arg]
            ),
            "def-uri-2": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter]
            ),
        },
    )

    with pytest.raises(
        errors.LabwareCannotBeStackedError, match="cannot be loaded to stack"
    ):
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(loadName="name"),  # type: ignore[call-arg]
                stackingOffsetWithLabware={"test": Vector3D(x=0, y=0, z=0)},
            ),
            bottom_labware_id="labware-id",
        )


@pytest.mark.parametrize(
    argnames=["primary_def", "lid_def", "adapter_def", "exception", "ordered_list"],
    argvalues=[
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"primary": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            does_not_raise(),
            ["lid", "primary", "adapter"],
            id="all-valid-and-present",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            None,
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            does_not_raise(),
            ["primary", "adapter"],
            id="adapter-valid-and-present",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"primary": Vector3D(x=0, y=0, z=0)},
            ),
            None,
            does_not_raise(),
            ["lid", "primary"],
            id="lid-valid-and-present",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            None,
            None,
            does_not_raise(),
            ["primary"],
            id="primary-only",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"uhoh": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            pytest.raises(errors.LabwareCannotBeStackedError),
            None,
            id="lid-may-not-stack-on-primary",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"uhoh": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"primary": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            pytest.raises(errors.LabwareCannotBeStackedError),
            None,
            id="primary-may-not-stack-on-adapter",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"primary": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            pytest.raises(errors.LabwareCannotBeStackedError),
            None,
            id="adapter-wrong-role",
        ),
        pytest.param(
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="primary"
                ),
                stackingOffsetWithLabware={"adapter": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="lid"
                ),
                stackingOffsetWithLabware={"primary": Vector3D(x=0, y=0, z=0)},
            ),
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(  # type: ignore[call-arg]
                    loadName="adapter"
                ),
            ),
            pytest.raises(errors.LabwareCannotBeStackedError),
            None,
            id="lid-wrong-role",
        ),
    ],
)
def test_stacker_labware_pool_passes_or_raises(
    primary_def: LabwareDefinition,
    lid_def: LabwareDefinition | None,
    adapter_def: LabwareDefinition | None,
    exception: ContextManager[None],
    ordered_list: List[str] | None,
) -> None:
    """It should raise if a stacker labware pool configuration is invalid."""
    subject = get_labware_view()
    with exception:
        subject.raise_if_stacker_labware_pool_is_not_valid(
            primary_def, lid_def, adapter_def
        )
    if exception is does_not_raise():
        result = subject.stacker_labware_pool_to_ordered_list(
            primary_def, lid_def, adapter_def
        )
        result_load_names = [labware_def.parameters.loadName for labware_def in result]
        assert result_load_names == ordered_list


@pytest.mark.parametrize(
    argnames=[
        "allowed_roles",
        "stack_limit",
        "exception",
    ],
    argvalues=[
        [
            [LabwareRole.labware],
            1,
            pytest.raises(errors.LabwareCannotBeStackedError),
        ],
        [
            [LabwareRole.lid],
            5,
            does_not_raise(),
        ],
    ],
)
def test_labware_stacking_height_passes_or_raises(
    allowed_roles: List[LabwareRole],
    stack_limit: int,
    exception: ContextManager[None],
) -> None:
    """It should raise if the labware is stacked too high, and pass if the labware definition allowed this."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id4": LoadedLabware(
                id="labware-id4",
                loadName="test",
                definitionUri="def-uri-1",
                location=OnLabwareLocation(labwareId="labware-id3"),
            ),
            "labware-id3": LoadedLabware(
                id="labware-id3",
                loadName="test",
                definitionUri="def-uri-1",
                location=OnLabwareLocation(labwareId="labware-id2"),
            ),
            "labware-id2": LoadedLabware(
                id="labware-id2",
                loadName="test",
                definitionUri="def-uri-1",
                location=OnLabwareLocation(labwareId="labware-id1"),
            ),
            "labware-id1": LoadedLabware(
                id="labware-id1",
                loadName="test",
                definitionUri="def-uri-1",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
        },
        definitions_by_uri={
            "def-uri-1": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=allowed_roles,
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="name",
                    isMagneticModuleCompatible=False,
                ),
                stackLimit=stack_limit,
            )
        },
    )

    with exception:
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="name",
                    isMagneticModuleCompatible=False,
                ),
                stackingOffsetWithLabware={"test": Vector3D(x=0, y=0, z=0)},
                stackLimit=stack_limit,
            ),
            bottom_labware_id="labware-id4",
        )


def test_tiprack_lid_stacking_height_passes_and_fails() -> None:
    """It should pass if the tiprack lid is on an adapter tiprack into lid stack-up, and fail if you attempt to stack multiple tiprack lids."""
    subject = get_labware_view(
        labware_by_id={
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="tiprack",
                definitionUri="def-uri-2",
                location=OnLabwareLocation(labwareId="adapter-id"),
            ),
            "adapter-id": LoadedLabware(
                id="adapter-id",
                loadName="tiprack_adapter",
                definitionUri="def-uri-1",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
        },
        definitions_by_uri={
            "def-uri-1": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="tiprack_adapter",
                    isMagneticModuleCompatible=False,
                ),
                stackLimit=1,
            ),
            "def-uri-2": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="tiprack",
                    isMagneticModuleCompatible=False,
                ),
                stackLimit=1,
            ),
        },
    )

    result = subject.raise_if_labware_cannot_be_stacked(
        top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            allowedRoles=[LabwareRole.lid],
            parameters=Parameters2.model_construct(
                format="irregular",
                isTiprack=False,
                loadName="opentrons_flex_tiprack_lid",
                isMagneticModuleCompatible=False,
            ),
            stackingOffsetWithLabware={"tiprack": Vector3D(x=0, y=0, z=0)},
            stackLimit=1,
        ),
        bottom_labware_id="labware-id",
    )
    assert result


def test_tiprack_lid_stacking_height_fails() -> None:
    """It should fail if the tiprack lid is stacked on itself higher than its max stacking height."""
    subject = get_labware_view(
        labware_by_id={
            "lid-id": LoadedLabware(
                id="lid-id",
                loadName="opentrons_flex_tiprack_lid",
                definitionUri="def-uri-1",
                location=OnLabwareLocation(labwareId="labware-id"),
            ),
            "labware-id": LoadedLabware(
                id="labware-id",
                loadName="tiprack",
                definitionUri="def-uri-2",
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
        },
        definitions_by_uri={
            "def-uri-1": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="opentrons_flex_tiprack_lid",
                    isMagneticModuleCompatible=False,
                ),
                stackLimit=1,
            ),
            "def-uri-2": LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="tiprack",
                    isMagneticModuleCompatible=False,
                ),
                stackLimit=1,
            ),
        },
    )
    with pytest.raises(errors.LabwareCannotBeStackedError, match="Tip rack lid"):
        subject.raise_if_labware_cannot_be_stacked(
            top_labware_definition=LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.lid],
                parameters=Parameters2.model_construct(
                    format="irregular",
                    isTiprack=False,
                    loadName="opentrons_flex_tiprack_lid",
                    isMagneticModuleCompatible=False,
                ),
                stackingOffsetWithLabware={
                    "opentrons_flex_tiprack_lid": Vector3D(x=0, y=0, z=0)
                },
                stackLimit=1,
            ),
            bottom_labware_id="lid-id",
        )


def test_get_grip_force(
    flex_50uL_tiprack: LabwareDefinition,
    reservoir_def: LabwareDefinition,
) -> None:
    """It should get the grip force, if present, from labware definition or return default."""
    subject = get_labware_view()

    assert subject.get_grip_force(flex_50uL_tiprack) == 16  # from definition
    assert subject.get_grip_force(reservoir_def) == 15  # default


def test_get_grip_z() -> None:
    """It should get the grip height, if present, from labware definition or return default."""
    subject = get_labware_view()

    schema_2_with_defined_height = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        schemaVersion=2, gripHeightFromLabwareBottom=123
    )
    assert subject.get_grip_z(schema_2_with_defined_height) == 123

    schema_3_with_defined_height = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        schemaVersion=3, gripHeightFromLabwareOrigin=123
    )
    assert subject.get_grip_z(schema_3_with_defined_height) == 123

    schema_2_without_defined_height = LabwareDefinition2.model_construct(  # type: ignore[call-arg]
        schemaVersion=2,
        dimensions=LabwareDimensions(xDimension=0, yDimension=0, zDimension=500),
    )
    assert subject.get_grip_z(schema_2_without_defined_height) == 250

    schema_3_without_defined_height = LabwareDefinition3.model_construct(  # type: ignore[call-arg]
        schemaVersion=3,
        extents=Extents(
            total=AxisAlignedBoundingBox3D(
                backLeftBottom=Vector3D(x=0, y=0, z=500),
                frontRightTop=Vector3D(x=0, y=0, z=1000),
            )
        ),
    )
    assert subject.get_grip_z(schema_3_without_defined_height) == 750


@pytest.mark.parametrize(
    "labware_to_check,well_bbox",
    [
        ("opentrons_universal_flat_adapter", Dimensions(0, 0, 0)),
        (
            "corning_96_wellplate_360ul_flat",
            Dimensions(116.81 - 10.95, 77.67 - 7.81, 14.22),
        ),
        ("nest_12_reservoir_15ml", Dimensions(117.48 - 10.28, 78.38 - 7.18, 31.4)),
    ],
)
def test_calculates_well_bounding_box(
    labware_to_check: str, well_bbox: Dimensions
) -> None:
    """It should be able to calculate well bounding boxes."""
    definition = labware_definition_type_adapter.validate_python(
        load_definition(labware_to_check, 1)
    )
    subject = get_labware_view()
    assert subject.get_well_bbox(definition).x == pytest.approx(well_bbox.x)
    assert subject.get_well_bbox(definition).y == pytest.approx(well_bbox.y)
    assert subject.get_well_bbox(definition).z == pytest.approx(well_bbox.z)


@pytest.mark.parametrize(
    "labware_to_check,gripper_specs",
    [
        (
            "opentrons_universal_flat_adapter",
            GripSpecs(targetY=75, uncertaintyNarrower=5, uncertaintyWider=0),
        ),
        # well min: 7.81
        # well max: 77.67
        # well bbox: 69.86
        (
            "corning_96_wellplate_360ul_flat",
            GripSpecs(targetY=85.47, uncertaintyNarrower=15.61, uncertaintyWider=0),
        ),
        # well min 7.18
        # well max 78.38
        # well bbox 71.2
        (
            "nest_12_reservoir_15ml",
            GripSpecs(targetY=85.48, uncertaintyNarrower=14.28, uncertaintyWider=0),
        ),
        (
            "opentrons_tough_universal_lid",
            GripSpecs(targetY=85.48, uncertaintyNarrower=5, uncertaintyWider=0),
        ),
        (
            "opentrons_flex_tiprack_lid",
            GripSpecs(targetY=78.75, uncertaintyNarrower=5, uncertaintyWider=0),
        ),
        # well min 7.175
        # well max 78.305
        # well bbox 71.13
        (
            "corning_384_wellplate_112ul_flat",
            GripSpecs(targetY=85.47, uncertaintyNarrower=14.34, uncertaintyWider=0),
        ),
    ],
)
def test_calculates_gripper_positions(
    labware_to_check: str, gripper_specs: GripSpecs
) -> None:
    """It should calculate gripper positions."""
    definition = labware_definition_type_adapter.validate_python(
        load_definition(labware_to_check, 1)
    )
    subject = get_labware_view()
    specs = subject.get_gripper_width_specs(definition)
    assert isclose(specs.targetY, gripper_specs.targetY)
    assert isclose(specs.uncertaintyNarrower, gripper_specs.uncertaintyNarrower)
    assert isclose(specs.uncertaintyWider, gripper_specs.uncertaintyWider)



"""Command factories to use in tests as data fixtures."""

from datetime import datetime
from typing import Dict, Optional, cast

from pydantic import BaseModel

from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.protocol_engine import ErrorOccurrence
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.types import (
    AddressableOffsetVector,
    DeckPoint,
    DeckSlotLocation,
    LabwareMovementStrategy,
    LiquidHandlingWellLocation,
    LoadableLabwareLocation,
    ModuleDefinition,
    ModuleModel,
    MovementAxis,
    WellLocation,
)
from opentrons.types import MountType


class FixtureModel(BaseModel):
    """Fixture Model."""

    ...


def create_queued_command(
    command_id: str = "command-id",
    command_key: str = "command-key",
    command_type: str = "command-type",
    intent: cmd.CommandIntent = cmd.CommandIntent.PROTOCOL,
    params: Optional[BaseModel] = None,
) -> cmd.Command:
    """Given command data, build a pending command model."""

    class DummyParams(BaseModel):
        pass

    return cast(
        cmd.Command,
        cmd.BaseCommand(
            id=command_id,
            key=command_key,
            commandType=command_type,
            createdAt=datetime(year=2021, month=1, day=1),
            status=cmd.CommandStatus.QUEUED,
            params=params or DummyParams(),
            intent=intent,
        ),
    )


def create_running_command(
    command_id: str = "command-id",
    command_key: str = "command-key",
    command_type: str = "command-type",
    created_at: datetime = datetime(year=2021, month=1, day=1),
    params: Optional[BaseModel] = None,
) -> cmd.Command:
    """Given command data, build a running command model."""
    return cast(
        cmd.Command,
        cmd.BaseCommand(
            id=command_id,
            key=command_key,
            createdAt=created_at,
            commandType=command_type,
            status=cmd.CommandStatus.RUNNING,
            params=params or FixtureModel(),
        ),
    )


def create_failed_command(
    command_id: str = "command-id",
    command_key: str = "command-key",
    command_type: str = "command-type",
    created_at: datetime = datetime(year=2021, month=1, day=1),
    completed_at: datetime = datetime(year=2022, month=2, day=2),
    params: Optional[BaseModel] = None,
    error: Optional[ErrorOccurrence] = None,
    intent: Optional[cmd.CommandIntent] = None,
) -> cmd.Command:
    """Given command data, build a failed command model."""
    return cast(
        cmd.Command,
        cmd.BaseCommand(
            id=command_id,
            key=command_key,
            createdAt=created_at,
            completedAt=completed_at,
            commandType=command_type,
            status=cmd.CommandStatus.FAILED,
            params=params or FixtureModel(),
            error=error,
            intent=intent,
        ),
    )


def create_succeeded_command(
    command_id: str = "command-id",
    command_key: str = "command-key",
    command_type: str = "command-type",
    created_at: datetime = datetime(year=2021, month=1, day=1),
    params: Optional[BaseModel] = None,
    result: Optional[BaseModel] = None,
) -> cmd.Command:
    """Given command data and result, build a completed command model."""
    return cast(
        cmd.Command,
        cmd.BaseCommand(
            id=command_id,
            key=command_key,
            createdAt=created_at,
            commandType=command_type,
            status=cmd.CommandStatus.SUCCEEDED,
            params=params or FixtureModel(),
            result=result or FixtureModel(),
        ),
    )


def create_comment_command(command_id: str = "command-id") -> cmd.Comment:
    """Create a completed LoadLabware command."""
    params = cmd.CommentParams(message="hello world!")

    result = cmd.CommentResult()

    return cmd.Comment(
        id=command_id,
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_load_pipette_command(
    pipette_id: str,
    pipette_name: PipetteNameType,
    mount: MountType,
) -> cmd.LoadPipette:
    """Get a completed LoadPipette command."""
    params = cmd.LoadPipetteParams(pipetteName=pipette_name, mount=mount)
    result = cmd.LoadPipetteResult(pipetteId=pipette_id)

    return cmd.LoadPipette(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_load_module_command(
    module_id: str,
    location: DeckSlotLocation,
    model: ModuleModel,
) -> cmd.LoadModule:
    """Get a completed LoadModule command."""
    params = cmd.LoadModuleParams(moduleId=module_id, location=location, model=model)
    result = cmd.LoadModuleResult(
        moduleId=module_id,
        model=model,
        serialNumber=None,
        definition=ModuleDefinition.model_construct(),  # type: ignore[call-arg]
    )

    return cmd.LoadModule(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_aspirate_command(
    pipette_id: str,
    volume: float,
    flow_rate: float,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    well_location: Optional[LiquidHandlingWellLocation] = None,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.Aspirate:
    """Get a completed Aspirate command."""
    params = cmd.AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location or LiquidHandlingWellLocation(),
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.AspirateResult(volume=volume, position=destination)

    return cmd.Aspirate(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_aspirate_while_tracking_command(
    pipette_id: str, volume: float, flow_rate: float, labware_id: str, well_name: str
) -> cmd.AspirateWhileTracking:
    """Get a completed Aspirate command."""
    params = cmd.AspirateWhileTrackingParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.AspirateWhileTrackingResult(volume=volume)

    return cmd.AspirateWhileTracking(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_aspirate_in_place_command(
    pipette_id: str,
    volume: float,
    flow_rate: float,
) -> cmd.AspirateInPlace:
    """Get a completed Aspirate command."""
    params = cmd.AspirateInPlaceParams(
        pipetteId=pipette_id,
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.AspirateInPlaceResult(volume=volume)

    return cmd.AspirateInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_dispense_command(
    pipette_id: str,
    volume: float,
    flow_rate: float,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    well_location: Optional[LiquidHandlingWellLocation] = None,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.Dispense:
    """Get a completed Dispense command."""
    params = cmd.DispenseParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location or LiquidHandlingWellLocation(),
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.DispenseResult(volume=volume, position=destination)

    return cmd.Dispense(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_dispense_in_place_command(
    pipette_id: str,
    volume: float,
    flow_rate: float,
) -> cmd.DispenseInPlace:
    """Get a completed DispenseInPlace command."""
    params = cmd.DispenseInPlaceParams(
        pipetteId=pipette_id,
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.DispenseInPlaceResult(volume=volume)

    return cmd.DispenseInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_dispense_while_tracking_command(
    pipette_id: str, volume: float, flow_rate: float, labware_id: str, well_name: str
) -> cmd.DispenseWhileTracking:
    """Get a completed DispenseWhileTracking command."""
    params = cmd.DispenseWhileTrackingParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        volume=volume,
        flowRate=flow_rate,
    )
    result = cmd.DispenseWhileTrackingResult(volume=volume)

    return cmd.DispenseWhileTracking(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_liquid_probe_command(
    pipette_id: str = "pippete-id",
    labware_id: str = "labware-id",
    well_name: str = "well-name",
    well_location: Optional[WellLocation] = None,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.LiquidProbe:
    """Get a completed Liquid Probe command."""
    params = cmd.LiquidProbeParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location or WellLocation(),
    )
    result = cmd.LiquidProbeResult(position=destination, z_position=0.5)

    return cmd.LiquidProbe(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_load_liquid_command(
    liquid_id: str = "liquid-id",
    labware_id: str = "labware-id",
    volume_by_well: Dict[str, float] = {"A1": 30, "B2": 100},
) -> cmd.LoadLiquid:
    """Get a completed Load Liquid command."""
    params = cmd.LoadLiquidParams(
        liquidId=liquid_id,
        labwareId=labware_id,
        volumeByWell=volume_by_well,
    )
    result = cmd.LoadLiquidResult()

    return cmd.LoadLiquid(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_pick_up_tip_command(
    pipette_id: str,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    tip_volume: float = 123.4,
    tip_length: float = 567.8,
    tip_diameter: float = 9.0,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.PickUpTip:
    """Get a completed PickUpTip command."""
    data = cmd.PickUpTipParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
    )

    result = cmd.PickUpTipResult(
        tipVolume=tip_volume,
        tipLength=tip_length,
        tipDiameter=tip_diameter,
        position=destination,
    )

    return cmd.PickUpTip(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=data,
        result=result,
    )


def create_drop_tip_command(
    pipette_id: str,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.DropTip:
    """Get a completed DropTip command."""
    params = cmd.DropTipParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
    )

    result = cmd.DropTipResult(position=destination)

    return cmd.DropTip(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_drop_tip_in_place_command(
    pipette_id: str,
) -> cmd.DropTipInPlace:
    """Get a completed DropTip command."""
    params = cmd.DropTipInPlaceParams(pipetteId=pipette_id)

    result = cmd.DropTipInPlaceResult()

    return cmd.DropTipInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_move_to_well_command(
    pipette_id: str,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.MoveToWell:
    """Get a completed MoveToWell command."""
    params = cmd.MoveToWellParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
    )

    result = cmd.MoveToWellResult(position=destination)

    return cmd.MoveToWell(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_move_to_addressable_area_command(
    pipette_id: str,
    addressable_area_name: str = "area-name",
    offset: AddressableOffsetVector = AddressableOffsetVector(x=0, y=0, z=0),
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.MoveToAddressableArea:
    """Get a completed MoveToWell command."""
    params = cmd.MoveToAddressableAreaParams(
        pipetteId=pipette_id,
        addressableAreaName=addressable_area_name,
        offset=offset,
    )

    result = cmd.MoveToAddressableAreaResult(position=destination)

    return cmd.MoveToAddressableArea(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_move_to_coordinates_command(
    pipette_id: str,
    coordinates: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.MoveToCoordinates:
    """Get a completed MoveToWell command."""
    params = cmd.MoveToCoordinatesParams(
        pipetteId=pipette_id,
        coordinates=coordinates,
    )

    result = cmd.MoveToCoordinatesResult(position=coordinates)

    return cmd.MoveToCoordinates(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_move_relative_command(
    pipette_id: str,
    axis: MovementAxis = MovementAxis.X,
    distance: float = 42,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.MoveRelative:
    """Get a completed MoveToWell command."""
    params = cmd.MoveRelativeParams(pipetteId=pipette_id, axis=axis, distance=distance)

    result = cmd.MoveRelativeResult(position=destination)

    return cmd.MoveRelative(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_blow_out_command(
    pipette_id: str,
    flow_rate: float,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    well_location: Optional[WellLocation] = None,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.BlowOut:
    """Get a completed BlowOut command."""
    params = cmd.BlowOutParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location or WellLocation(),
        flowRate=flow_rate,
    )
    result = cmd.BlowOutResult(position=destination)

    return cmd.BlowOut(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime(year=2022, month=1, day=1),
        params=params,
        result=result,
    )


def create_blow_out_in_place_command(
    pipette_id: str,
    flow_rate: float,
) -> cmd.BlowOutInPlace:
    """Get a completed blowOutInPlace command."""
    params = cmd.BlowOutInPlaceParams(
        pipetteId=pipette_id,
        flowRate=flow_rate,
    )
    result = cmd.BlowOutInPlaceResult()

    return cmd.BlowOutInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime(year=2022, month=1, day=1),
        params=params,
        result=result,
    )


def create_touch_tip_command(
    pipette_id: str,
    labware_id: str = "labware-id",
    well_name: str = "A1",
    well_location: Optional[WellLocation] = None,
    radius: float = 1.0,
    speed: Optional[float] = None,
    destination: DeckPoint = DeckPoint(x=0, y=0, z=0),
) -> cmd.TouchTip:
    """Get a completed BlowOut command."""
    params = cmd.TouchTipParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location or WellLocation(),
        radius=radius,
        speed=speed,
    )
    result = cmd.TouchTipResult(position=destination)

    return cmd.TouchTip(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime(year=2022, month=1, day=1),
        params=params,
        result=result,
    )


def create_move_labware_command(
    new_location: LoadableLabwareLocation,
    strategy: LabwareMovementStrategy,
    labware_id: str = "labware-id",
    offset_id: Optional[str] = None,
) -> cmd.MoveLabware:
    """Get a completed MoveLabware command."""
    params = cmd.MoveLabwareParams(
        labwareId=labware_id,
        newLocation=new_location,
        strategy=strategy,
    )
    result = cmd.MoveLabwareResult(offsetId=offset_id)

    return cmd.MoveLabware(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime(year=2022, month=1, day=1),
        params=params,
        result=result,
    )


def create_prepare_to_aspirate_command(pipette_id: str) -> cmd.PrepareToAspirate:
    """Get a completed PrepareToAspirate command."""
    params = cmd.PrepareToAspirateParams(pipetteId=pipette_id)
    result = cmd.PrepareToAspirateResult()
    return cmd.PrepareToAspirate(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime(year=2023, month=10, day=24),
        params=params,
        result=result,
    )


def create_reload_labware_command(
    labware_id: str,
    offset_id: Optional[str],
) -> cmd.ReloadLabware:
    """Create a completed ReloadLabware command."""
    params = cmd.ReloadLabwareParams(
        labwareId=labware_id,
    )

    result = cmd.ReloadLabwareResult(
        labwareId=labware_id,
        offsetId=offset_id,
    )

    return cmd.ReloadLabware(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_unsafe_blow_out_in_place_command(
    pipette_id: str,
    flow_rate: float,
) -> cmd.unsafe.UnsafeBlowOutInPlace:
    """Create a completed UnsafeBlowOutInPlace command."""
    params = cmd.unsafe.UnsafeBlowOutInPlaceParams(
        pipetteId=pipette_id, flowRate=flow_rate
    )
    result = cmd.unsafe.UnsafeBlowOutInPlaceResult()

    return cmd.unsafe.UnsafeBlowOutInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )


def create_unsafe_drop_tip_in_place_command(
    pipette_id: str,
) -> cmd.unsafe.UnsafeDropTipInPlace:
    """Get a completed UnsafeDropTipInPlace command."""
    params = cmd.unsafe.UnsafeDropTipInPlaceParams(pipetteId=pipette_id)

    result = cmd.unsafe.UnsafeDropTipInPlaceResult()

    return cmd.unsafe.UnsafeDropTipInPlace(
        id="command-id",
        key="command-key",
        status=cmd.CommandStatus.SUCCEEDED,
        createdAt=datetime.now(),
        params=params,
        result=result,
    )



"""Tests for the PipetteStore+PipetteState+PipetteView trifecta.

The trifecta is tested here as a single unit, treating PipetteState as a private
implementation detail.
"""

from collections import OrderedDict

import pytest

from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

from ..pipette_fixtures import (
    EIGHT_CHANNEL_COLS,
    EIGHT_CHANNEL_MAP,
    EIGHT_CHANNEL_ROWS,
    NINETY_SIX_COLS,
    NINETY_SIX_MAP,
    NINETY_SIX_ROWS,
)
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine import actions, commands
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.pipettes import PipetteStore, PipetteView
from opentrons.protocol_engine.types import FlowRates, LabwareWellId, TipGeometry
from opentrons.types import MountType, Point


def _dummy_command() -> commands.Command:
    """Return a placeholder command."""
    return commands.Comment.model_construct()  # type: ignore[call-arg]


def test_handle_pipette_config_action(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
) -> None:
    """Should add pipette channel to state."""
    subject = PipetteStore()

    config_update = update_types.PipetteConfigUpdate(
        pipette_id="pipette-id",
        serial_number="pipette-serial",
        config=LoadedStaticPipetteData(
            channels=8,
            max_volume=15,
            min_volume=3,
            model="gen a",
            display_name="display name",
            flow_rates=FlowRates(
                default_aspirate={},
                default_dispense={},
                default_blow_out={},
            ),
            tip_configuration_lookup_table={15: supported_tip_fixture},
            nominal_tip_overlap={},
            nozzle_offset_z=1.23,
            home_position=4.56,
            nozzle_map=NozzleMap.build(
                physical_nozzles=EIGHT_CHANNEL_MAP,
                physical_rows=EIGHT_CHANNEL_ROWS,
                physical_columns=EIGHT_CHANNEL_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H1",
                valid_nozzle_maps=pipette_definition.ValidNozzleMaps(
                    maps={"Full": EIGHT_CHANNEL_COLS["1"]}
                ),
            ),
            back_left_corner_offset=Point(x=1, y=2, z=3),
            front_right_corner_offset=Point(x=4, y=5, z=6),
            pipette_lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=pipette_definition.AvailableSensorDefinition(
                sensors=["pressure", "capacitive", "environment"]
            ),
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        ),
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(pipette_config=config_update),
            command=_dummy_command(),
        )
    )

    assert PipetteView(subject.state).get_channels("pipette-id") == 8
    assert PipetteView(subject.state).get_active_channels("pipette-id") == 8


@pytest.mark.parametrize(
    argnames=["nozzle_map", "expected_channels"],
    argvalues=[
        (
            NozzleMap.build(
                physical_nozzles=OrderedDict({"A1": Point(0, 0, 0)}),
                physical_rows=OrderedDict({"A": ["A1"]}),
                physical_columns=OrderedDict({"1": ["A1"]}),
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="A1",
                valid_nozzle_maps=pipette_definition.ValidNozzleMaps(
                    maps={"A1": ["A1"]}
                ),
            ),
            1,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="H12",
                valid_nozzle_maps=pipette_definition.ValidNozzleMaps(
                    maps={
                        "Full": sum(
                            [
                                NINETY_SIX_ROWS["A"],
                                NINETY_SIX_ROWS["B"],
                                NINETY_SIX_ROWS["C"],
                                NINETY_SIX_ROWS["D"],
                                NINETY_SIX_ROWS["E"],
                                NINETY_SIX_ROWS["F"],
                                NINETY_SIX_ROWS["G"],
                                NINETY_SIX_ROWS["H"],
                            ],
                            [],
                        )
                    }
                ),
            ),
            96,
        ),
        (
            NozzleMap.build(
                physical_nozzles=NINETY_SIX_MAP,
                physical_rows=NINETY_SIX_ROWS,
                physical_columns=NINETY_SIX_COLS,
                starting_nozzle="A1",
                back_left_nozzle="A1",
                front_right_nozzle="E1",
                valid_nozzle_maps=pipette_definition.ValidNozzleMaps(
                    maps={"A1_E1": ["A1", "B1", "C1", "D1", "E1"]}
                ),
            ),
            5,
        ),
    ],
)
def test_active_channels(
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    nozzle_map: NozzleMap,
    expected_channels: int,
) -> None:
    """Should update active channels after pipette configuration change."""
    subject = PipetteStore()

    # Load pipette to update state
    config_update = update_types.PipetteConfigUpdate(
        pipette_id="pipette-id",
        serial_number="pipette-serial",
        config=LoadedStaticPipetteData(
            channels=9,
            max_volume=15,
            min_volume=3,
            model="gen a",
            display_name="display name",
            flow_rates=FlowRates(
                default_aspirate={},
                default_dispense={},
                default_blow_out={},
            ),
            tip_configuration_lookup_table={15: supported_tip_fixture},
            nominal_tip_overlap={},
            nozzle_offset_z=1.23,
            home_position=4.56,
            nozzle_map=nozzle_map,
            back_left_corner_offset=Point(x=1, y=2, z=3),
            front_right_corner_offset=Point(x=4, y=5, z=6),
            pipette_lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=pipette_definition.AvailableSensorDefinition(
                sensors=["pressure", "capacitive", "environment"]
            ),
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        ),
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(pipette_config=config_update),
            command=_dummy_command(),
        )
    )

    # Configure nozzle for partial configuration
    state_update = update_types.StateUpdate(
        pipette_nozzle_map=update_types.PipetteNozzleMapUpdate(
            pipette_id="pipette-id",
            nozzle_map=nozzle_map,
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=_dummy_command(),
            state_update=state_update,
        )
    )
    assert (
        PipetteView(subject.state).get_active_channels("pipette-id")
        == expected_channels
    )


def test_last_tip_rack_well() -> None:
    """Should update last tip rack well after pipette load and tip state updates."""
    subject = PipetteStore()

    load_pipette_update = update_types.LoadPipetteUpdate(
        pipette_id="pipette-id",
        pipette_name=PipetteNameType.P50_SINGLE_FLEX,
        mount=MountType.RIGHT,
        liquid_presence_detection=None,
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(loaded_pipette=load_pipette_update),
            command=_dummy_command(),
        )
    )

    assert (
        PipetteView(subject.state).get_tip_rack_well_picked_up_from("pipette-id")
        is None
    )

    tip_picked_up_update = update_types.PipetteTipStateUpdate(
        pipette_id="pipette-id",
        tip_geometry=TipGeometry(length=1, diameter=2, volume=3),
        tip_source=LabwareWellId(
            labware_id="my-cool-labware", well_name="less-cool-well"
        ),
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(
                pipette_tip_state=tip_picked_up_update
            ),
            command=_dummy_command(),
        )
    )

    assert PipetteView(subject.state).get_tip_rack_well_picked_up_from(
        "pipette-id"
    ) == LabwareWellId(labware_id="my-cool-labware", well_name="less-cool-well")

    tip_dropped_update = update_types.PipetteTipStateUpdate(
        pipette_id="pipette-id", tip_geometry=None, tip_source=None
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            state_update=update_types.StateUpdate(pipette_tip_state=tip_dropped_update),
            command=_dummy_command(),
        )
    )

    assert (
        PipetteView(subject.state).get_tip_rack_well_picked_up_from("pipette-id")
        is None
    )



"""Tests for the ParameterContext public interface."""

import inspect

import pytest
from decoy import Decoy

from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
)
from opentrons.protocol_api._parameter_context import ParameterContext
from opentrons.protocol_engine.types import BooleanParameter
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parameters import (
    csv_parameter_definition as mock_csv_parameter_definition,
)
from opentrons.protocols.parameters import (
    parameter_definition as mock_parameter_definition,
)
from opentrons.protocols.parameters import (
    validation as mock_validation,
)
from opentrons.protocols.parameters.exceptions import (
    IncompatibleParameterError,
    ParameterDefinitionError,
)


@pytest.fixture(autouse=True)
def _mock_parameter_definition_creates(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    for name, func in inspect.getmembers(mock_parameter_definition, inspect.isfunction):
        monkeypatch.setattr(mock_parameter_definition, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_csv_parameter_definition_creates(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    for name, func in inspect.getmembers(
        mock_csv_parameter_definition, inspect.isfunction
    ):
        monkeypatch.setattr(mock_csv_parameter_definition, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _patch_parameter_validation(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture
def api_version() -> APIVersion:
    """The API version under test."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(api_version: APIVersion) -> ParameterContext:
    """Get a ParameterContext test subject."""
    return ParameterContext(api_version=api_version)


def test_add_int(decoy: Decoy, subject: ParameterContext) -> None:
    """It should create and add an int parameter definition."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_parameter_definition.ParameterDefinition
    )
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def.variable_name).then_return("my cool variable")
    decoy.when(
        mock_parameter_definition.create_int_parameter(
            display_name="abc",
            variable_name="xyz",
            default=123,
            minimum=45,
            maximum=678,
            choices=[{"display_name": "foo", "value": 42}],
            description="blah blah blah",
            unit="foot candles",
        )
    ).then_return(param_def)

    subject.add_int(
        display_name="abc",
        variable_name="xyz",
        default=123,
        minimum=45,
        maximum=678,
        choices=[{"display_name": "foo", "value": 42}],
        description="blah blah blah",
        unit="foot candles",
    )

    assert param_def is subject._parameters["my cool variable"]
    decoy.verify(mock_validation.validate_variable_name_unique("xyz", {"other_param"}))


def test_add_float(decoy: Decoy, subject: ParameterContext) -> None:
    """It should create and add a float parameter definition."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_parameter_definition.ParameterDefinition
    )
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def.variable_name).then_return("my cooler variable")
    decoy.when(mock_validation.ensure_float_value(12.3)).then_return(3.21)
    decoy.when(mock_validation.ensure_optional_float_value(4.5)).then_return(5.4)
    decoy.when(mock_validation.ensure_optional_float_value(67.8)).then_return(87.6)
    decoy.when(
        mock_validation.ensure_float_choices([{"display_name": "foo", "value": 4.2}])
    ).then_return([{"display_name": "bar", "value": 2.4}])
    decoy.when(
        mock_parameter_definition.create_float_parameter(
            display_name="abc",
            variable_name="xyz",
            default=3.21,
            minimum=5.4,
            maximum=87.6,
            choices=[{"display_name": "bar", "value": 2.4}],
            description="blah blah blah",
            unit="lux",
        )
    ).then_return(param_def)

    subject.add_float(
        display_name="abc",
        variable_name="xyz",
        default=12.3,
        minimum=4.5,
        maximum=67.8,
        choices=[{"display_name": "foo", "value": 4.2}],
        description="blah blah blah",
        unit="lux",
    )

    assert param_def is subject._parameters["my cooler variable"]
    decoy.verify(mock_validation.validate_variable_name_unique("xyz", {"other_param"}))


def test_add_bool(decoy: Decoy, subject: ParameterContext) -> None:
    """It should create and add a boolean parameter definition."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_parameter_definition.ParameterDefinition
    )
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def.variable_name).then_return("my coolest variable")
    decoy.when(
        mock_parameter_definition.create_bool_parameter(
            display_name="cba",
            variable_name="zxy",
            default=False,
            choices=[
                {"display_name": "On", "value": True},
                {"display_name": "Off", "value": False},
            ],
            description="lorem ipsum",
        )
    ).then_return(param_def)

    subject.add_bool(
        display_name="cba",
        variable_name="zxy",
        default=False,
        description="lorem ipsum",
    )

    assert param_def is subject._parameters["my coolest variable"]
    decoy.verify(mock_validation.validate_variable_name_unique("zxy", {"other_param"}))


def test_add_string(decoy: Decoy, subject: ParameterContext) -> None:
    """It should create and add a string parameter definition."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_parameter_definition.ParameterDefinition
    )
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def.variable_name).then_return("my slightly less cool variable")
    decoy.when(
        mock_parameter_definition.create_str_parameter(
            display_name="jkl",
            variable_name="qwerty",
            default="asdf",
            choices=[{"display_name": "bar", "value": "aaa"}],
            description="fee foo fum",
        )
    ).then_return(param_def)

    subject.add_str(
        display_name="jkl",
        variable_name="qwerty",
        default="asdf",
        choices=[{"display_name": "bar", "value": "aaa"}],
        description="fee foo fum",
    )

    assert param_def is subject._parameters["my slightly less cool variable"]
    decoy.verify(
        mock_validation.validate_variable_name_unique("qwerty", {"other_param"})
    )


def test_add_csv(decoy: Decoy, subject: ParameterContext) -> None:
    """It should create and add a CSV parameter definition."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_parameter_definition.ParameterDefinition
    )
    param_def = decoy.mock(cls=mock_csv_parameter_definition.CSVParameterDefinition)
    decoy.when(param_def.variable_name).then_return("my potentially cool variable")
    decoy.when(
        mock_csv_parameter_definition.create_csv_parameter(
            display_name="jkl",
            variable_name="qwerty",
            description="fee foo fum",
        )
    ).then_return(param_def)

    subject.add_csv_file(
        display_name="jkl",
        variable_name="qwerty",
        description="fee foo fum",
    )

    assert param_def is subject._parameters["my potentially cool variable"]
    decoy.verify(
        mock_validation.validate_variable_name_unique("qwerty", {"other_param"})
    )


def test_add_csv_raises_for_multiple_csvs(
    decoy: Decoy, subject: ParameterContext
) -> None:
    """It should raise with a IncompatibleParameterError if there's already a CSV parameter.."""
    subject._parameters["other_param"] = decoy.mock(
        cls=mock_csv_parameter_definition.CSVParameterDefinition
    )

    with pytest.raises(IncompatibleParameterError):
        subject.add_csv_file(
            display_name="jkl",
            variable_name="qwerty",
            description="fee foo fum",
        )


def test_set_parameters(decoy: Decoy, subject: ParameterContext) -> None:
    """It should set the parameter values."""
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def.parameter_type).then_return(bool)
    decoy.when(mock_validation.ensure_value_type("bar", bool)).then_return("rhubarb")
    subject._parameters["foo"] = param_def

    subject.set_parameters({"foo": "bar"})

    assert param_def.value == "rhubarb"


def test_set_parameters_raises(decoy: Decoy, subject: ParameterContext) -> None:
    """It should raise if the given parameter is not defined."""
    with pytest.raises(ParameterDefinitionError):
        subject.set_parameters({"foo": "bar"})


def test_export_parameters_for_analysis(
    decoy: Decoy, subject: ParameterContext
) -> None:
    """It should export the parameters as protocol engine types."""
    param_def = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    boolean_param = decoy.mock(cls=BooleanParameter)
    decoy.when(param_def.as_protocol_engine_type()).then_return(boolean_param)
    subject._parameters["foo"] = param_def

    assert subject.export_parameters_for_analysis() == [boolean_param]


def test_export_parameters_for_protocol(
    decoy: Decoy, subject: ParameterContext
) -> None:
    """It should export the parameters as a Parameters object with the parameters as dynamic attributes."""
    param_def_1 = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    param_def_2 = decoy.mock(cls=mock_parameter_definition.ParameterDefinition)
    decoy.when(param_def_1.variable_name).then_return("x")
    decoy.when(param_def_1.value).then_return("a")
    decoy.when(param_def_2.variable_name).then_return("y")
    decoy.when(param_def_2.value).then_return(1.23)
    subject._parameters = {"foo": param_def_1, "bar": param_def_2}

    result = subject.export_parameters_for_protocol()
    assert result.x == "a"  # type: ignore[attr-defined]
    assert result.y == 1.23  # type: ignore[attr-defined]



"""Test save position command."""

import pytest
from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.save_position import (
    SavePositionImplementation,
    SavePositionParams,
    SavePositionResult,
)
from opentrons.protocol_engine.execution import GantryMover
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.types import DeckPoint
from opentrons.types import Point


@pytest.fixture
def mock_model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def mock_gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mock GantryMover."""
    return decoy.mock(cls=GantryMover)


async def test_save_position_implementation(
    decoy: Decoy,
    mock_model_utils: ModelUtils,
    mock_gantry_mover: GantryMover,
) -> None:
    """A SavePosition command should have an execution implementation."""
    subject = SavePositionImplementation(
        model_utils=mock_model_utils, gantry_mover=mock_gantry_mover
    )
    params = SavePositionParams(pipetteId="abc", positionId="123", failOnNotHomed=True)

    decoy.when(mock_model_utils.ensure_id("123")).then_return("456")

    decoy.when(
        await mock_gantry_mover.get_position(pipette_id="abc", fail_on_not_homed=True)
    ).then_return(Point(x=1, y=2, z=3))

    result = await subject.execute(params)

    assert result == SuccessData(
        public=SavePositionResult(
            positionId="456",
            position=DeckPoint(x=1, y=2, z=3),
        ),
    )



import asyncio
from typing import AsyncIterator

import pytest

from opentrons.hardware_control.emulation.proxy import (
    Proxy,
    ProxyListener,
)
from opentrons.hardware_control.emulation.settings import ProxySettings


@pytest.fixture
def setting() -> ProxySettings:
    """Proxy settings fixture."""
    return ProxySettings(emulator_port=12345, driver_port=12346)


class SimpleProxyListener(ProxyListener):
    def __init__(self) -> None:
        self._count = 0

    def on_server_connected(
        self, server_type: str, client_uri: str, identifier: str
    ) -> None:
        self._count += 1

    def on_server_disconnected(self, identifier: str) -> None:
        self._count -= 1

    async def wait_count(self, count: int) -> None:
        while count != self._count:
            await asyncio.sleep(0.01)


@pytest.fixture
def proxy_listener() -> SimpleProxyListener:
    """A proxy listener."""
    return SimpleProxyListener()


@pytest.fixture
async def subject(
    setting: ProxySettings,
    proxy_listener: SimpleProxyListener,
) -> AsyncIterator[Proxy]:
    """Test subject."""
    p = Proxy("proxy", proxy_listener, setting)
    task = asyncio.get_running_loop().create_task(p.run())
    await p.wait_ready()
    yield p
    p.stop()

    try:
        await task
    except asyncio.CancelledError:
        pass


async def test_driver_route_message(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should route a message to an emulator."""
    emulator = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(1)
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    driver[1].write(b"abc\n")
    r = await emulator[0].readline()
    assert r == b"abc\n"


async def test_emulator_route_message(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should route a message to a driver."""
    emulator = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(1)
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)

    test_data = b"abc\n"
    emulator[1].write(test_data)
    r = await driver[0].readline()
    assert r == test_data


async def test_driver_route_message_two_connections(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should route messages to correct emulator."""
    emulator1 = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    emulator2 = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(2)
    driver1 = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    driver2 = await asyncio.open_connection(host="localhost", port=setting.driver_port)

    test_data1 = b"abc\n"
    test_data2 = b"cba\n"
    driver1[1].write(test_data1)
    driver2[1].write(test_data2)
    r1 = await emulator1[0].readline()
    r2 = await emulator2[0].readline()
    assert r1 == test_data1
    assert r2 == test_data2


async def test_emulator_route_message_two_connections(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should route messages to correct driver."""
    emulator1 = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    emulator2 = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(2)
    driver1 = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    driver2 = await asyncio.open_connection(host="localhost", port=setting.driver_port)

    test_data1 = b"abc\n"
    test_data2 = b"cba\n"
    emulator1[1].write(test_data1)
    emulator2[1].write(test_data2)
    r1 = await driver1[0].readline()
    r2 = await driver2[0].readline()
    assert r1 == test_data1
    assert r2 == test_data2


async def test_driver_and_no_emulator(subject: Proxy, setting: ProxySettings) -> None:
    """It should fail to read if no emulator."""
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    assert b"" == await driver[0].readline()


async def test_two_driver_and_one_emulator(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should fail to read if no emulator."""
    emulator = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(1)
    driver1 = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    driver2 = await asyncio.open_connection(host="localhost", port=setting.driver_port)

    test_data1 = b"abc\n"
    emulator[1].write(test_data1)
    assert test_data1 == await driver1[0].readline()
    assert b"" == await driver2[0].readline()


async def test_driver_reconnect(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should allow a second driver to claim a formerly used emulator."""
    emulator = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(1)
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)

    test_data = b"abc\n"
    emulator[1].write(test_data)
    assert test_data == await driver[0].readline()

    driver[1].close()
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    emulator[1].write(test_data)
    assert test_data == await driver[0].readline()


async def test_emulator_disconnects(
    subject: Proxy, setting: ProxySettings, proxy_listener: SimpleProxyListener
) -> None:
    """It should disconnect driver when emulator disconnects."""
    emulator = await asyncio.open_connection(
        host="localhost", port=setting.emulator_port
    )
    await proxy_listener.wait_count(1)
    driver = await asyncio.open_connection(host="localhost", port=setting.driver_port)
    emulator[1].close()
    driver[1].write(b"123\n")
    assert b"" == await driver[0].readline()



"""Functional tests for the LabwareDataProvider."""

from typing import cast

from opentrons_shared_data.labware.labware_definition import (
    labware_definition_type_adapter,
)
from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict

from opentrons.calibration_storage.helpers import hash_labware_def
from opentrons.protocol_api.labware import get_labware_definition
from opentrons.protocol_engine.resources import LabwareDataProvider


async def test_labware_data_gets_standard_definition() -> None:
    """It should be able to get a "standard" labware's definition."""
    expected = get_labware_definition(
        load_name="opentrons_96_tiprack_300ul",
        namespace="opentrons",
        version=1,
    )
    result = await LabwareDataProvider().get_labware_definition(
        load_name="opentrons_96_tiprack_300ul",
        namespace="opentrons",
        version=1,
    )

    assert result == labware_definition_type_adapter.validate_python(expected)


async def test_labware_hash_match() -> None:
    """Labware dict vs Pydantic model hashing should match.

    This is a smoke test to ensure proper functioning of
    LabwareDataProvider.get_calibrated_tip_length given its
    internal implementation relies on this to work.
    """
    labware_dict = get_labware_definition(
        load_name="opentrons_96_tiprack_300ul",
        namespace="opentrons",
        version=1,
    )

    labware_model = labware_definition_type_adapter.validate_python(labware_dict)
    labware_model_dict = cast(
        LabwareDefDict, labware_model.model_dump(exclude_none=True, exclude_unset=True)
    )

    assert hash_labware_def(labware_dict) == hash_labware_def(labware_model_dict)



"""Common absorbance reader defines."""

# This needs to be re-defined because of circular imports
_PLATE_READER_MAX_LABWARE_Z_MM = 16.0
# Distance to home the gripper when moving the plate reader lid to/from the dock.
LID_Z_CLEARANCE = _PLATE_READER_MAX_LABWARE_Z_MM + 40.0



"""Test calibrate-module command."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

import pytest
from decoy import Decoy

from opentrons.hardware_control import ot3_calibration as calibration
from opentrons.hardware_control.api import API as OT2API
from opentrons.hardware_control.types import OT3Mount
from opentrons.protocol_engine.commands.calibration.calibrate_module import (
    CalibrateModuleImplementation,
    CalibrateModuleParams,
    CalibrateModuleResult,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors.exceptions import HardwareNotSupportedError
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    ModuleOffsetData,
    ModuleOffsetVector,
)
from opentrons.types import DeckSlotName, MountType, Point

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture(autouse=True)
def _mock_ot3_calibration(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(calibration, inspect.isfunction):
        monkeypatch.setattr(calibration, name, decoy.mock(func=func))


@pytest.mark.ot3_only
async def test_calibrate_module_implementation(
    decoy: Decoy, ot3_hardware_api: OT3API, state_view: StateView
) -> None:
    """Test Calibration command execution."""
    subject = CalibrateModuleImplementation(state_view, ot3_hardware_api)

    location = DeckSlotLocation(slotName=DeckSlotName("D3"))
    module_id = "module123"
    labware_id = "labware123"
    module_serial = "TC1234abcd"
    params = CalibrateModuleParams(
        moduleId=module_id,
        labwareId=labware_id,
        mount=MountType.LEFT,
    )

    decoy.when(state_view.modules.get_serial_number(module_id)).then_return(
        "TC1234abcd"
    )

    decoy.when(state_view.modules.get_location(module_id)).then_return(location)
    decoy.when(state_view.modules.get_module_calibration_offset(module_id)).then_return(
        ModuleOffsetData(
            moduleOffsetVector=ModuleOffsetVector(x=0, y=0, z=0),
            location=location,
        )
    )
    decoy.when(
        state_view.geometry.get_well_position(labware_id=labware_id, well_name="B1")
    ).then_return(Point(x=3, y=2, z=1))
    decoy.when(
        await calibration.calibrate_module(
            hcapi=ot3_hardware_api,
            mount=OT3Mount.LEFT,
            slot=location.slotName.id,
            module_id=module_serial,
            nominal_position=Point(x=3, y=2, z=1),
        )
    ).then_return(Point(x=3, y=4, z=6))

    result = await subject.execute(params)

    assert result == SuccessData(
        public=CalibrateModuleResult(
            moduleOffset=ModuleOffsetVector(
                x=3,
                y=4,
                z=6,
            ),
            location=location,
        ),
    )


@pytest.mark.ot3_only
async def test_calibrate_module_implementation_wrong_hardware(
    decoy: Decoy, ot2_hardware_api: OT2API, state_view: StateView
) -> None:
    """Should raise an unsupported hardware error."""
    subject = CalibrateModuleImplementation(
        state_view=state_view, hardware_api=ot2_hardware_api
    )

    params = CalibrateModuleParams(
        moduleId="Test1234",
        labwareId="Test1234",
        mount=MountType.LEFT,
    )

    with pytest.raises(HardwareNotSupportedError):
        await subject.execute(params)



"""Test get next tip in place commands."""

from decoy import Decoy

from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine import StateView
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.get_next_tip import (
    GetNextTipImplementation,
    GetNextTipParams,
    GetNextTipResult,
)
from opentrons.protocol_engine.types import NextTipInfo, NoTipAvailable, NoTipReason
from opentrons.types import NozzleConfigurationType


async def test_get_next_tip_implementation(
    decoy: Decoy,
    state_view: StateView,
) -> None:
    """A GetNextTip command should have an execution implementation."""
    subject = GetNextTipImplementation(state_view=state_view)
    params = GetNextTipParams(
        pipetteId="abc", labwareIds=["123"], startingTipWell="xyz"
    )
    mock_nozzle_map = decoy.mock(cls=NozzleMap)

    decoy.when(state_view.pipettes.get_active_channels("abc")).then_return(42)
    decoy.when(state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        mock_nozzle_map
    )
    decoy.when(mock_nozzle_map.configuration).then_return(NozzleConfigurationType.FULL)

    decoy.when(
        state_view.tips.get_next_tip(
            labware_id="123",
            num_tips=42,
            starting_tip_name="xyz",
            nozzle_map=mock_nozzle_map,
        )
    ).then_return("foo")

    result = await subject.execute(params)

    assert result == SuccessData(
        public=GetNextTipResult(
            nextTipInfo=NextTipInfo(labwareId="123", tipStartingWell="foo")
        ),
    )


async def test_get_next_tip_implementation_multiple_tip_racks(
    decoy: Decoy,
    state_view: StateView,
) -> None:
    """A GetNextTip command with multiple tip racks should not apply starting tip to the following ones."""
    subject = GetNextTipImplementation(state_view=state_view)
    params = GetNextTipParams(
        pipetteId="abc", labwareIds=["123", "456"], startingTipWell="xyz"
    )
    mock_nozzle_map = decoy.mock(cls=NozzleMap)

    decoy.when(state_view.pipettes.get_active_channels("abc")).then_return(42)
    decoy.when(state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        mock_nozzle_map
    )
    decoy.when(mock_nozzle_map.configuration).then_return(NozzleConfigurationType.FULL)

    decoy.when(
        state_view.tips.get_next_tip(
            labware_id="456",
            num_tips=42,
            starting_tip_name=None,
            nozzle_map=mock_nozzle_map,
        )
    ).then_return("foo")

    result = await subject.execute(params)

    assert result == SuccessData(
        public=GetNextTipResult(
            nextTipInfo=NextTipInfo(labwareId="456", tipStartingWell="foo")
        ),
    )


async def test_get_next_tip_implementation_no_tips(
    decoy: Decoy,
    state_view: StateView,
) -> None:
    """A GetNextTip command should return with NoTipAvailable if there are no available tips."""
    subject = GetNextTipImplementation(state_view=state_view)
    params = GetNextTipParams(
        pipetteId="abc", labwareIds=["123", "456"], startingTipWell="xyz"
    )
    mock_nozzle_map = decoy.mock(cls=NozzleMap)

    decoy.when(state_view.pipettes.get_active_channels("abc")).then_return(42)
    decoy.when(state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        mock_nozzle_map
    )
    decoy.when(mock_nozzle_map.configuration).then_return(NozzleConfigurationType.FULL)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=GetNextTipResult(
            nextTipInfo=NoTipAvailable(
                noTipReason=NoTipReason.NO_AVAILABLE_TIPS,
                message="No available tips for given pipette, nozzle configuration and provided tip racks.",
            )
        ),
    )


async def test_get_next_tip_implementation_partial_with_starting_tip(
    decoy: Decoy,
    state_view: StateView,
) -> None:
    """A GetNextTip command should return with NoTipAvailable if there's a starting tip and a partial config."""
    subject = GetNextTipImplementation(state_view=state_view)
    params = GetNextTipParams(
        pipetteId="abc", labwareIds=["123", "456"], startingTipWell="xyz"
    )
    mock_nozzle_map = decoy.mock(cls=NozzleMap)

    decoy.when(state_view.pipettes.get_active_channels("abc")).then_return(42)
    decoy.when(state_view.pipettes.get_nozzle_configuration("abc")).then_return(
        mock_nozzle_map
    )
    decoy.when(mock_nozzle_map.configuration).then_return(NozzleConfigurationType.ROW)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=GetNextTipResult(
            nextTipInfo=NoTipAvailable(
                noTipReason=NoTipReason.STARTING_TIP_WITH_PARTIAL,
                message="Cannot automatically resolve next tip with starting tip and partial tip configuration.",
            )
        ),
    )



"""Fixtures for protocol engine command tests."""

from __future__ import annotations

import pytest
from decoy import Decoy

from opentrons.protocol_engine.actions import ActionDispatcher
from opentrons.protocol_engine.execution import (
    EquipmentHandler,
    GantryMover,
    LabwareMovementHandler,
    MovementHandler,
    PipettingHandler,
    RailLightsHandler,
    RunControlHandler,
    StatusBarHandler,
    TaskHandler,
    TipHandler,
)
from opentrons.protocol_engine.resources import (
    ConcurrencyProvider,
    FileProvider,
    ModelUtils,
)
from opentrons.protocol_engine.state.state import StateStore, StateView


@pytest.fixture
def state_view(decoy: Decoy) -> StateView:
    """Get a mocked out StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def equipment(decoy: Decoy) -> EquipmentHandler:
    """Get a mocked out EquipmentHandler."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def movement(decoy: Decoy) -> MovementHandler:
    """Get a mocked out MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def labware_movement(decoy: Decoy) -> LabwareMovementHandler:
    """Get a mocked out LabwareMovementHandler."""
    return decoy.mock(cls=LabwareMovementHandler)


@pytest.fixture
def pipetting(decoy: Decoy) -> PipettingHandler:
    """Get a mocked out PipettingHandler."""
    return decoy.mock(cls=PipettingHandler)


@pytest.fixture
def tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mocked out EquipmentHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.fixture
def run_control(decoy: Decoy) -> RunControlHandler:
    """Get a mocked out RunControlHandler."""
    return decoy.mock(cls=RunControlHandler)


@pytest.fixture
def rail_lights(decoy: Decoy) -> RailLightsHandler:
    """Get a mocked out RailLightsHandler."""
    return decoy.mock(cls=RailLightsHandler)


@pytest.fixture
def model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mocked out ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def status_bar(decoy: Decoy) -> StatusBarHandler:
    """Get a mocked out StatusBarHandler."""
    return decoy.mock(cls=StatusBarHandler)


@pytest.fixture
def gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mocked out GantryMover."""
    return decoy.mock(cls=GantryMover)


@pytest.fixture
def file_provider(decoy: Decoy) -> FileProvider:
    """Get a mocked out StateView."""
    return decoy.mock(cls=FileProvider)


@pytest.fixture
def task_handler(decoy: Decoy) -> TaskHandler:
    """Get a mocked out TaskHandler."""
    return decoy.mock(cls=TaskHandler)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mocked out ActionDispatcher.

    Only use this if you are using a real task handler.
    """
    return decoy.mock(cls=ActionDispatcher)


@pytest.fixture
def concurrency_provider(decoy: Decoy) -> ConcurrencyProvider:
    """Get a mocked out ConcurrencyProvider.

    Only use this if you are using a real task handler.
    """
    return decoy.mock(cls=ConcurrencyProvider)


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore.

    Only use this if you are using a real task handler.
    """
    return decoy.mock(cls=StateStore)


@pytest.fixture
def real_concurrency_provider() -> ConcurrencyProvider:
    """Get a real concurrency provider."""
    return ConcurrencyProvider()


@pytest.fixture
def real_task_handler(
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    real_concurrency_provider: ConcurrencyProvider,
) -> TaskHandler:
    """Get a real task handler with mocked dependencies."""
    return TaskHandler(
        state_store=state_store,
        action_dispatcher=action_dispatcher,
        model_utils=model_utils,
        concurrency_provider=real_concurrency_provider,
    )



import pytest

from opentrons_shared_data.gripper.gripper_definition import GripperModel

from opentrons.config import gripper_config as gc


def test_load_gripper_config() -> None:
    loaded_config = gc.load(GripperModel.v1)
    assert loaded_config.display_name == "Flex Gripper"


def test_gripper_force_conversion() -> None:
    loaded_config = gc.load(GripperModel.v1)
    for force in range(0, 25):
        if 5 <= force <= 20:
            expected = 2.09 * force - 0.282
            result = gc.duty_cycle_by_force(force, loaded_config.grip_force_profile)
            assert result == expected
        else:
            with pytest.raises(ValueError):
                gc.duty_cycle_by_force(force, loaded_config.grip_force_profile)



"""Test Heater Shaker set shake speed command implementation."""

import pytest
from decoy import Decoy, matchers

from opentrons.hardware_control.modules import HeaterShaker
from opentrons.protocol_engine.actions import Action, ActionDispatcher, StartTaskAction
from opentrons.protocol_engine.commands import heater_shaker
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.heater_shaker.set_shake_speed import (
    SetShakeSpeedImpl,
)
from opentrons.protocol_engine.execution import (
    EquipmentHandler,
    MovementHandler,
    TaskHandler,
)
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.module_substates import (
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import MotorAxis
from opentrons.protocol_engine.types.tasks import Task


@pytest.mark.parametrize(
    ("pipette_blocking_hs_shaker", "expect_pipette_retracted"),
    [
        (False, False),
        (True, True),
    ],
)
async def test_set_shake_speed(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    movement: MovementHandler,
    pipette_blocking_hs_shaker: bool,
    expect_pipette_retracted: bool,
) -> None:
    """It should be able to set the module's shake speed."""
    subject = SetShakeSpeedImpl(
        state_view=state_view,
        equipment=equipment,
        movement=movement,
        task_handler=real_task_handler,
    )
    data = heater_shaker.SetShakeSpeedParams(
        moduleId="input-heater-shaker-id", rpm=1234.56, taskId="taskId"
    )

    hs_module_substate = decoy.mock(cls=HeaterShakerModuleSubState)
    hs_hardware = decoy.mock(cls=HeaterShaker)

    decoy.when(
        state_view.modules.get_heater_shaker_module_substate(
            module_id="input-heater-shaker-id"
        )
    ).then_return(hs_module_substate)

    decoy.when(hs_module_substate.module_id).then_return(
        HeaterShakerModuleId("heater-shaker-id")
    )
    decoy.when(model_utils.ensure_id("taskId")).then_return("taskId")
    decoy.when(
        state_view.motion.check_pipette_blocking_hs_shaker(
            HeaterShakerModuleId("heater-shaker-id")
        )
    ).then_return(pipette_blocking_hs_shaker)

    # Stub speed validation from hs module view
    decoy.when(hs_module_substate.validate_target_speed(rpm=1234.56)).then_return(1234)

    # Get attached hardware modules
    decoy.when(
        equipment.get_module_hardware_api(HeaterShakerModuleId("heater-shaker-id"))
    ).then_return(hs_hardware)
    decoy.when(state_view.motion.get_robot_mount_axes()).then_return(
        [MotorAxis.EXTENSION_Z]
    )
    task: Task | None = None

    def _capture_task(action: Action) -> None:
        nonlocal task
        assert isinstance(action, StartTaskAction)
        task = action.task

    decoy.when(
        action_dispatcher.dispatch(StartTaskAction(task=matchers.Anything()))  # type: ignore[func-returns-value]
    ).then_do(_capture_task)
    result = await subject.execute(data)
    assert task is not None
    await task.asyncioTask

    decoy.verify(hs_module_substate.raise_if_labware_latch_not_closed())
    if expect_pipette_retracted:
        decoy.verify(await movement.home([MotorAxis.EXTENSION_Z]))
    decoy.verify(await hs_hardware.set_speed(rpm=1234))

    assert result == SuccessData(
        public=heater_shaker.SetShakeSpeedResult(
            pipetteRetracted=expect_pipette_retracted, taskId="taskId"
        ),
        state_update=(
            update_types.StateUpdate(pipette_location=update_types.CLEAR)
            if expect_pipette_retracted
            else update_types.StateUpdate()
        ),
    )



import asyncio
from typing import AsyncGenerator

import pytest
from decoy import Decoy

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.vacuum_module.simulator import SimulatingDriver
from opentrons.drivers.vacuum_module.types import (
    HardwareRevision,
    LEDColor,
    LEDPattern,
    VacuumModuleInfo,
)
from opentrons.hardware_control import ExecutionManager, modules
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
)
from opentrons.hardware_control.modules.vacuum_module import (
    SIMULATING_POLL_PERIOD,
    VacuumModuleReader,
)
from opentrons.hardware_control.poller import Poller
from opentrons.hardware_control.types import StatusBarState, StatusBarUpdateEvent


@pytest.fixture
def usb_port() -> USBPort:
    """Token USB port."""
    return USBPort(
        name="",
        port_number=0,
        device_path="/dev/ot_module_sim_vacuummodule0",
    )


@pytest.fixture
def mock_driver(decoy: Decoy) -> SimulatingDriver:
    """Mocked simulating driver."""
    return decoy.mock(cls=SimulatingDriver)


@pytest.fixture
async def subject(
    usb_port: USBPort,
    mock_driver: SimulatingDriver,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
    decoy: Decoy,
) -> AsyncGenerator[modules.VacuumModule, None]:
    """Test subject with mocked driver."""
    reader = VacuumModuleReader(driver=mock_driver)
    poller = Poller(reader=reader, interval=SIMULATING_POLL_PERIOD)
    vacuum = modules.VacuumModule(
        port="/dev/ot_module_sim_vacuummodule0",
        usb_port=usb_port,
        driver=mock_driver,
        reader=reader,
        poller=poller,
        device_info={
            "serial": "dummySerialFS",
            "model": "a1",
            "version": "vacuum-fw",
        },
        hw_control_loop=asyncio.get_running_loop(),
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    decoy.when(await mock_driver.get_device_info()).then_return(
        VacuumModuleInfo(fw="vacuum-fw", hw=HardwareRevision.NFF, sn="dummySerialFS")
    )

    await poller.start()
    try:
        yield vacuum
    finally:
        await vacuum.cleanup()


async def test_sim_state(subject: modules.VacuumModule) -> None:
    """It should forward state."""
    status = subject.device_info
    assert status["serial"] == "dummySerialFS"
    assert status["model"] == "a1"
    assert status["version"] == "vacuum-fw"


@pytest.mark.parametrize(
    ("should_identify", "event", "result_params"),
    [
        (  # running
            False,
            StatusBarUpdateEvent(state=StatusBarState.RUNNING, enabled=True),
            (0.5, LEDColor.GREEN, LEDPattern.STATIC, None),
        ),
        (  # paused - should identify
            True,
            StatusBarUpdateEvent(state=StatusBarState.PAUSED, enabled=True),
            (0.5, LEDColor.BLUE, LEDPattern.PULSE, 2000),
        ),
        (  # paused - door closed not identified
            False,
            StatusBarUpdateEvent(state=StatusBarState.PAUSED, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # idle - door closed
            False,
            StatusBarUpdateEvent(state=StatusBarState.IDLE, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # hardware error - identified
            True,
            StatusBarUpdateEvent(state=StatusBarState.HARDWARE_ERROR, enabled=True),
            (0.5, LEDColor.RED, LEDPattern.FLASH, 300),
        ),
        (  # hardware error - not identified
            False,
            StatusBarUpdateEvent(state=StatusBarState.HARDWARE_ERROR, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # software error
            False,
            StatusBarUpdateEvent(state=StatusBarState.SOFTWARE_ERROR, enabled=True),
            (0.5, LEDColor.YELLOW, LEDPattern.STATIC, None),
        ),
        (  # error recovery - should identify
            True,
            StatusBarUpdateEvent(state=StatusBarState.ERROR_RECOVERY, enabled=True),
            (0.5, LEDColor.YELLOW, LEDPattern.PULSE, 2000),
        ),
        (  # error recovery - door closed
            False,
            StatusBarUpdateEvent(state=StatusBarState.ERROR_RECOVERY, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # run complete
            False,
            StatusBarUpdateEvent(state=StatusBarState.RUN_COMPLETED, enabled=True),
            (0.5, LEDColor.GREEN, LEDPattern.PULSE, None),
        ),
        (  # updating
            False,
            StatusBarUpdateEvent(state=StatusBarState.UPDATING, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.PULSE, None),
        ),
    ],
)
async def test_statusbar_event_handler(
    subject: modules.VacuumModule,
    mock_driver: SimulatingDriver,
    should_identify: bool,
    event: StatusBarUpdateEvent,
    result_params: tuple[float, LEDColor, LEDPattern, int | None],
    decoy: Decoy,
) -> None:
    """It should handle LED lights."""
    subject.set_statusbar_identify(should_identify)
    await subject._handle_status_bar_event(event)
    decoy.verify(
        await mock_driver.set_led(
            result_params[0],
            color=result_params[1],
            pattern=result_params[2],
            duration=result_params[3],
            reps=None,
        )
    )
